#!/usr/bin/python3

"""This file is part of asyncoro; see http://asyncoro.sourceforge.net for
details.

This program can be used to start discoro server processes so discoro scheduler
(see 'discoro.py') can send computations to these server processes for executing
distributed communicating proceses (coroutines). All coroutines in a server
execute in the same thread, so multiple CPUs are not used by one server. If CPU
intensive computations are to be run on systems with multiple processors, then
this program should be run with multiple instances (see below for '-c' option to
this program).

See 'discoro_client*.py' files for example use cases.
"""

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2014 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"


def _discoro_server_coro_proc():
    # coroutine
    """Server process receives computations and runs coroutines for it.
    """

    import os
    import shutil
    import traceback
    import sys
    import time

    from asyncoro.discoro import MinPulseInterval, MaxPulseInterval, \
        DiscoroNodeInfo, DiscoroNodeAvailInfo, Scheduler
    import asyncoro.disasyncoro as asyncoro
    from asyncoro.disasyncoro import Coro, SysCoro, Location

    _discoro_coro = asyncoro.AsynCoro.cur_coro()
    _discoro_config = yield _discoro_coro.receive()
    _discoro_node_coro = asyncoro.deserialize(_discoro_config['node_coro'])
    _discoro_scheduler_coro = asyncoro.deserialize(_discoro_config['scheduler_coro'])
    assert isinstance(_discoro_scheduler_coro, Coro)
    _discoro_computation_auth = _discoro_config.pop('computation_auth', None)

    if _discoro_config['min_pulse_interval'] > 0:
        MinPulseInterval = _discoro_config['min_pulse_interval']
    if _discoro_config['max_pulse_interval'] > 0:
        MaxPulseInterval = _discoro_config['max_pulse_interval']
    _discoro_busy_time = _discoro_config.pop('busy_time')
    asyncoro.MsgTimeout = _discoro_config.pop('msg_timeout')

    _discoro_name = asyncoro.AsynCoro.instance().name
    _discoro_dest_path = os.path.join(asyncoro.AsynCoro.instance().dest_path,
                                      'discoroproc-%s' % _discoro_config['id'])
    if os.path.isdir(_discoro_dest_path):
        shutil.rmtree(_discoro_dest_path)
    asyncoro.AsynCoro.instance().dest_path = _discoro_dest_path
    os.chdir(_discoro_dest_path)
    sys.path.insert(0, _discoro_dest_path)

    for _discoro_var in _discoro_config.pop('peers', []):
        Coro(asyncoro.AsynCoro.instance().peer, asyncoro.deserialize(_discoro_var))

    for _discoro_var in ['clean', 'min_pulse_interval', 'max_pulse_interval']:
        del _discoro_config[_discoro_var]

    _discoro_coro.register('discoro_server')
    asyncoro.logger.info('discoro server %s started at %s; '
                         'computation files will be saved in "%s"',
                         _discoro_config['id'], _discoro_coro.location, _discoro_dest_path)
    _discoro_req = _discoro_client = _discoro_auth = _discoro_msg = None
    _discoro_peer_status = _discoro_monitor_coro = _discoro_monitor_proc = _discoro_job = None
    _discoro_job_coros = set()
    _discoro_jobs_done = asyncoro.Event()

    def _discoro_peer_status(coro=None):
        coro.set_daemon()
        while 1:
            status = yield coro.receive()
            if not isinstance(status, asyncoro.PeerStatus):
                asyncoro.logger.warning('Invalid peer status %s ignored', type(status))
                continue
            if status.status == asyncoro.PeerStatus.Offline:
                if (_discoro_scheduler_coro and
                    _discoro_scheduler_coro.location == status.location):
                    if _discoro_computation_auth:
                        _discoro_coro.send({'req': 'close', 'auth': _discoro_computation_auth})

    def _discoro_monitor_proc(zombie_period, coro=None):
        coro.set_daemon()
        while 1:
            msg = yield coro.receive(timeout=zombie_period)
            if isinstance(msg, asyncoro.MonitorException):
                asyncoro.logger.debug('coro %s done', msg.args[0])
                _discoro_job_coros.discard(msg.args[0])
                if not _discoro_job_coros:
                    _discoro_jobs_done.set()
                _discoro_busy_time.value = int(time.time())
            elif not msg:
                if _discoro_job_coros:
                    _discoro_busy_time.value = int(time.time())
            else:
                asyncoro.logger.warning('invalid message to monitor ignored: %s', type(msg))

    _discoro_var = _discoro_config['computation_location']
    _discoro_var = asyncoro.Location(_discoro_var.addr, _discoro_var.port)
    if (yield asyncoro.AsynCoro.instance().peer(_discoro_var)):
        raise StopIteration(-1)
    asyncoro.AsynCoro.instance().peer_status(SysCoro(_discoro_peer_status))
    yield asyncoro.AsynCoro.instance().peer(_discoro_node_coro.location)
    yield asyncoro.AsynCoro.instance().peer(_discoro_scheduler_coro.location)
    _discoro_scheduler_coro.send({'status': Scheduler.ServerDiscovered,
                                  'coro': _discoro_coro, 'name': _discoro_name,
                                  'auth': _discoro_computation_auth})

    if _discoro_config['_server_setup']:
        if _discoro_config['_disable_servers']:
            while 1:
                _discoro_var = yield _discoro_coro.receive()
                if (isinstance(_discoro_var, dict) and
                    _discoro_var.get('req', None) == 'enable_server' and
                    _discoro_var.get('auth', None) == _discoro_computation_auth):
                    _discoro_var = _discoro_var['setup_args']
                    if not isinstance(_discoro_var, tuple):
                        _discoro_var = tuple(_discoro_var)
                    break
                else:
                    asyncoro.logger.warning('Ignoring invalid request to run server setup')
        else:
            _discoro_var = ()
        _discoro_var = yield asyncoro.Coro(globals()[_discoro_config['_server_setup']],
                                           *_discoro_var).finish()
        if _discoro_var:
            asyncoro.logger.debug('discoro server %s @ %s setup failed',
                                  _discoro_config['id'], _discoro_coro.location)
            raise StopIteration(_discoro_var)
        _discoro_config['_server_setup'] = None
    _discoro_scheduler_coro.send({'status': Scheduler.ServerInitialized,
                                  'coro': _discoro_coro, 'name': _discoro_name,
                                  'auth': _discoro_computation_auth})

    _discoro_var = _discoro_config['zombie_period']
    if _discoro_var:
        _discoro_var /= 3
    else:
        _discoro_var = None
    _discoro_monitor_coro = SysCoro(_discoro_monitor_proc, _discoro_var)
    _discoro_node_coro.send({'req': 'server_setup', 'id': _discoro_config['id'],
                             'coro': _discoro_coro})
    _discoro_busy_time.value = int(time.time())
    asyncoro.logger.debug('discoro server "%s": Computation "%s" from %s',
                          _discoro_name, _discoro_computation_auth,
                          _discoro_scheduler_coro.location)

    while 1:
        _discoro_msg = yield _discoro_coro.receive()
        if not isinstance(_discoro_msg, dict):
            continue
        _discoro_req = _discoro_msg.get('req', None)

        if _discoro_req == 'run':
            _discoro_client = _discoro_msg.get('client', None)
            _discoro_auth = _discoro_msg.get('auth', None)
            _discoro_job = _discoro_msg.get('job', None)
            if (not isinstance(_discoro_client, Coro) or
                _discoro_auth != _discoro_computation_auth):
                asyncoro.logger.warning('invalid run: %s', type(_discoro_job))
                if isinstance(_discoro_client, Coro):
                    _discoro_client.send(None)
                continue
            try:
                if _discoro_job.code:
                    exec(_discoro_job.code, globals())
                _discoro_job.args = asyncoro.deserialize(_discoro_job.args)
                _discoro_job.kwargs = asyncoro.deserialize(_discoro_job.kwargs)
            except:
                asyncoro.logger.debug('invalid computation to run')
                _discoro_var = (sys.exc_info()[0], _discoro_job.name, traceback.format_exc())
                _discoro_client.send(_discoro_var)
            else:
                Coro._asyncoro._lock.acquire()
                try:
                    _discoro_var = Coro(globals()[_discoro_job.name],
                                        *(_discoro_job.args), **(_discoro_job.kwargs))
                except:
                    _discoro_var = (sys.exc_info()[0], _discoro_job.name, traceback.format_exc())
                else:
                    _discoro_job_coros.add(_discoro_var)
                    _discoro_busy_time.value = int(time.time())
                    asyncoro.logger.debug('coro %s created', _discoro_var)
                    _discoro_var.notify(_discoro_monitor_coro)
                    _discoro_var.notify(_discoro_scheduler_coro)
                _discoro_client.send(_discoro_var)
                Coro._asyncoro._lock.release()

        elif _discoro_req == 'close' or _discoro_req == 'quit':
            _discoro_auth = _discoro_msg.get('auth', None)
            if (_discoro_auth == _discoro_computation_auth):
                pass
            elif (_discoro_msg.get('node_auth', None) == _discoro_config['node_auth']):
                if _discoro_scheduler_coro:
                    _discoro_scheduler_coro.send({'status': Scheduler.ServerClosed,
                                                  'location': _discoro_coro.location})
                while _discoro_job_coros:
                    asyncoro.logger.debug('discoro server "%s": Waiting for %s coroutines to '
                                          'terminate before closing computation',
                                          _discoro_name, len(_discoro_job_coros))
                    if (yield _discoro_jobs_done.wait(timeout=5)):
                        break
            else:
                continue
            _discoro_var = _discoro_msg.get('client', None)
            if isinstance(_discoro_var, Coro):
                _discoro_var.send(0)
            break

        elif _discoro_req == 'terminate':
            _discoro_auth = _discoro_msg.get('node_auth', None)
            if (_discoro_auth != _discoro_config['node_auth']):
                continue
            if _discoro_scheduler_coro:
                _discoro_scheduler_coro.send({'status': Scheduler.ServerDisconnected,
                                              'location': _discoro_coro.location})
            break

        elif _discoro_req == 'status':
            if _discoro_msg.get('node_auth', None) != _discoro_config['node_auth']:
                continue
            if _discoro_scheduler_coro:
                print('  discoro server "%s" @ %s with PID %s running %d coroutines for %s' %
                      (_discoro_name, _discoro_coro.location, os.getpid(),
                       len(_discoro_job_coros), _discoro_scheduler_coro.location))
            else:
                print('  discoro server "%s" @ %s with PID %s not used by any computation' %
                      (_discoro_name, _discoro_coro.location, os.getpid()))

        elif _discoro_req == 'peers':
            _discoro_auth = _discoro_msg.get('auth', None)
            if (_discoro_auth == _discoro_computation_auth):
                for _discoro_var in _discoro_msg.get('peers', []):
                    asyncoro.Coro(asyncoro.AsynCoro.instance().peer, _discoro_var)

        else:
            asyncoro.logger.warning('invalid command "%s" ignored', _discoro_req)
            _discoro_client = _discoro_msg.get('client', None)
            if not isinstance(_discoro_client, Coro):
                continue
            _discoro_client.send(-1)

    # kill any pending jobs
    while _discoro_job_coros:
        for _discoro_job_coro in _discoro_job_coros:
            _discoro_job_coro.terminate()
        asyncoro.logger.debug('discoro server "%s": Waiting for %s coroutines to terminate '
                              'before closing computation', _discoro_name, len(_discoro_job_coros))
        if (yield _discoro_jobs_done.wait(timeout=5)):
            break
    asyncoro.logger.debug('discoro server %s @ %s done',
                          _discoro_config['id'], _discoro_coro.location)


def _discoro_server_process(_discoro_config, _discoro_mp_queue, _discoro_computation):
    import os
    import sys
    import time
    # import traceback

    for _discoro_var in list(sys.modules.keys()):
        if _discoro_var.startswith('asyncoro'):
            sys.modules.pop(_discoro_var)
    globals().pop('asyncoro', None)

    global asyncoro
    import asyncoro.disasyncoro as asyncoro

    _discoro_pid_path = os.path.join(_discoro_config['dest_path'],
                                     'discoroproc-%s.pid' % _discoro_config['id'])
    if os.path.isfile(_discoro_pid_path):
        with open(_discoro_pid_path, 'r') as _discoro_req:
            _discoro_var = _discoro_req.read()
        _discoro_var = int(_discoro_var)
        if not _discoro_config['clean']:
            print('\n   Another discoronode seems to be running;\n'
                  '   make sure server with PID %d quit and remove "%s"\n' %
                  (_discoro_var, _discoro_pid_path))
            _discoro_var = os.getpid()

        import signal
        try:
            os.kill(_discoro_var, signal.SIGINT)
            time.sleep(0.1)
            os.kill(_discoro_var, signal.SIGKILL)
        except:
            pass
        else:
            time.sleep(0.1)
            try:
                if os.waitpid(_discoro_var, os.WNOHANG)[0] != _discoro_var:
                    asyncoro.logger.warning('Killing process %d failed', _discoro_var)
            except:
                pass
        del signal, _discoro_req

    with open(_discoro_pid_path, 'w') as _discoro_var:
        _discoro_var.write('%s' % os.getpid())

    if _discoro_config['loglevel']:
        asyncoro.logger.setLevel(asyncoro.logger.DEBUG)
        # asyncoro.logger.show_ms(True)
    else:
        asyncoro.logger.setLevel(asyncoro.logger.INFO)
    del _discoro_config['loglevel']

    server_id = _discoro_config['id']
    mp_queue, _discoro_mp_queue = _discoro_mp_queue, None
    config = {}
    for _discoro_var in ['udp_port', 'tcp_port', 'node', 'ext_ip_addr', 'name', 'discover_peers',
                         'secret', 'certfile', 'keyfile', 'dest_path', 'max_file_size']:
        config[_discoro_var] = _discoro_config.pop(_discoro_var, None)

    while 1:
        try:
            _discoro_scheduler = asyncoro.AsynCoro(**config)
        except:
            print('discoro server %s failed for port %s; retrying in 5 seconds' %
                  (server_id, config['tcp_port']))
            # print(traceback.format_exc())
            time.sleep(5)
        else:
            break

    if os.name == 'nt':
        _discoro_computation = asyncoro.deserialize(_discoro_computation)
        if _discoro_computation._code:
            exec(_discoro_computation._code, globals())
        if __name__ == '__mp_main__':  # Windows multiprocessing process
            sys.modules['__mp_main__'].__dict__.update(globals())

    _discoro_config['_disable_servers'] = _discoro_computation._disable_servers
    _discoro_config['_server_setup'] = _discoro_computation._server_setup
    _discoro_config['computation_location'] = _discoro_computation._pulse_coro.location
    _discoro_coro = asyncoro.SysCoro(_discoro_server_coro_proc)
    assert isinstance(_discoro_coro, asyncoro.Coro)
    mp_queue.put((server_id, asyncoro.serialize(_discoro_coro)))
    _discoro_coro.send(_discoro_config)

    _discoro_config = None
    del config, _discoro_var

    _discoro_coro.value()
    _discoro_scheduler.ignore_peers(ignore=True)
    for location in _discoro_scheduler.peers():
        asyncoro.Coro(_discoro_scheduler.close_peer, location)
    _discoro_scheduler.finish()
    try:
        os.remove(_discoro_pid_path)
    except:
        pass
    mp_queue.put((server_id, None))
    exit(0)


def _discoro_spawn(_discoro_config, _discoro_id_ports, _discoro_mp_queue,
                   _discoro_pipe, _discoro_computation, _discoro_setup_args):
    import os
    import sys
    import signal
    import multiprocessing
    # import traceback

    try:
        signal.signal(signal.SIGHUP, signal.SIG_DFL)
        signal.signal(signal.SIGQUIT, signal.SIG_DFL)
    except:
        pass
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)

    for _discoro_var in list(sys.modules.keys()):
        if _discoro_var.startswith('asyncoro'):
            sys.modules.pop(_discoro_var)
    globals().pop('asyncoro', None)

    import asyncoro

    os.chdir(_discoro_config['dest_path'])
    sys.path.insert(0, _discoro_config['dest_path'])
    os.environ['PATH'] = _discoro_config['dest_path'] + os.pathsep + os.environ['PATH']
    procs = [None] * len(_discoro_id_ports)

    def terminate(status):
        for i in range(len(_discoro_id_ports)):
            proc = procs[i]
            if not proc:
                continue
            if proc.is_alive():
                try:
                    proc.terminate()
                except:
                    pass
                else:
                    proc.join(1)
            if (not proc.is_alive()) and proc.exitcode:
                asyncoro.logger.warning('Server %s (process %s) reaped', _discoro_id_ports[i][0],
                                        proc.pid)
                _discoro_mp_queue.put((_discoro_id_ports[i][0], None))
            _discoro_pid_path = os.path.join(_discoro_config['dest_path'],
                                             'discoroproc-%s.pid' % _discoro_id_ports[i][0])
            try:
                os.remove(_discoro_pid_path)
            except:
                pass

        _discoro_pipe.send('closed')
        exit(status)

    if os.name != 'nt':
        if _discoro_computation._code:
            exec(_discoro_computation._code, globals())

        if _discoro_computation._node_setup:
            try:
                if not isinstance(_discoro_setup_args, tuple):
                    _discoro_setup_args = tuple(_discoro_setup_args)
                ret = asyncoro.Coro(globals()[_discoro_computation._node_setup],
                                    *_discoro_setup_args).value()
            except:
                asyncoro.logger.warning('node_setup failed for %s', _discoro_computation._auth)
                # print(traceback.format_exc())
                ret = -1
            if ret != 0:
                _discoro_pipe.send(0)
                terminate(ret)
            _discoro_computation._node_setup = None

    def start_process(i, procs):
        server_config = dict(_discoro_config)
        server_config['id'] = _discoro_id_ports[i][0]
        server_config['name'] = '%s_proc-%s' % (_discoro_config['name'], server_config['id'])
        server_config['tcp_port'] = _discoro_id_ports[i][1]
        server_config['peers'] = _discoro_config['peers'][:]
        procs[i] = multiprocessing.Process(target=_discoro_server_process,
                                           name=server_config['name'],
                                           args=(server_config, _discoro_mp_queue,
                                                 _discoro_computation))
        procs[i].start()

    for i in range(len(procs)):
        start_process(i, procs)
        asyncoro.logger.debug('discoro server %s started with PID %s',
                              _discoro_id_ports[i][0], procs[i].pid)

    _discoro_pipe.send(len(procs))

    while 1:
        req = _discoro_pipe.recv()
        if req['req'] == 'quit':
            break
        else:
            asyncoro.logger.warning('Ignoring invalid pipe cmd: %s' % str(req))

    for proc in procs:
        proc.join(1)

    terminate(0)


if __name__ == '__main__':

    """
    See http://asyncoro.sourceforge.net/discoro.html#node-servers for details on
    options to start this program.
    """

    import sys
    import time
    import argparse
    import multiprocessing
    import threading
    import socket
    import os
    import hashlib
    import re
    import signal
    import platform
    import shutil
    try:
        import readline
    except:
        pass
    try:
        import psutil
    except ImportError:
        print('\n   \'psutil\' module is not available; '
              'CPU, memory, disk status will not be sent!\n')
        psutil = None
    else:
        psutil.cpu_percent(0.1)
    from asyncoro.discoro import MinPulseInterval, MaxPulseInterval, Scheduler, Computation
    import asyncoro.disasyncoro as asyncoro

    parser = argparse.ArgumentParser()
    parser.add_argument('--config', dest='config', default='',
                        help='use configuration in given file')
    parser.add_argument('--save_config', dest='save_config', default='',
                        help='save configuration in given file and exit')
    parser.add_argument('-c', '--cpus', dest='cpus', type=int, default=0,
                        help='number of CPUs/discoro instances to run; '
                        'if negative, that many CPUs are not used')
    parser.add_argument('-i', '--ip_addr', dest='node', default='',
                        help='IP address or host name of this node')
    parser.add_argument('--ext_ip_addr', dest='ext_ip_addr', default='',
                        help='External IP address to use (needed in case of NAT firewall/gateway)')
    parser.add_argument('-u', '--udp_port', dest='udp_port', type=int, default=51351,
                        help='UDP port number to use')
    parser.add_argument('--tcp_ports', dest='tcp_ports', action='append', default=[],
                        help='TCP port numbers to use')
    parser.add_argument('--scheduler_port', dest='scheduler_port', type=int, default=51350,
                        help='UDP port number used by discoro scheduler')
    parser.add_argument('-n', '--name', dest='name', default='',
                        help='(symbolic) name given to AsynCoro schdulers on this node')
    parser.add_argument('--dest_path', dest='dest_path', default='',
                        help='path prefix to where files sent by peers are stored')
    parser.add_argument('--max_file_size', dest='max_file_size', default='',
                        help='maximum file size of any file transferred')
    parser.add_argument('-s', '--secret', dest='secret', default='',
                        help='authentication secret for handshake with peers')
    parser.add_argument('--certfile', dest='certfile', default='',
                        help='file containing SSL certificate')
    parser.add_argument('--keyfile', dest='keyfile', default='',
                        help='file containing SSL key')
    parser.add_argument('--serve', dest='serve', default=-1, type=int,
                        help='number of clients to serve before exiting')
    parser.add_argument('--service_start', dest='service_start', default='',
                        help='time of day in HH:MM format when to start service')
    parser.add_argument('--service_stop', dest='service_stop', default='',
                        help='time of day in HH:MM format when to stop service '
                        '(continue to execute running jobs, but no new jobs scheduled)')
    parser.add_argument('--service_end', dest='service_end', default='',
                        help='time of day in HH:MM format when to end service '
                        '(terminate running jobs)')
    parser.add_argument('--msg_timeout', dest='msg_timeout', default=asyncoro.MsgTimeout, type=int,
                        help='timeout for delivering messages')
    parser.add_argument('--min_pulse_interval', dest='min_pulse_interval',
                        default=MinPulseInterval, type=int,
                        help='minimum pulse interval clients can use in number of seconds')
    parser.add_argument('--max_pulse_interval', dest='max_pulse_interval',
                        default=MaxPulseInterval, type=int,
                        help='maximum pulse interval clients can use in number of seconds')
    parser.add_argument('--zombie_period', dest='zombie_period', default=(10 * MaxPulseInterval),
                        type=int, help='maximum number of seconds for client to not run computation')
    parser.add_argument('--ping_interval', dest='ping_interval', default=0, type=int,
                        help='interval in number of seconds for node to broadcast its address')
    parser.add_argument('--daemon', action='store_true', dest='daemon', default=False,
                        help='if given, input is not read from terminal')
    parser.add_argument('--clean', action='store_true', dest='clean', default=False,
                        help='if given, server processes from previous run will be killed '
                        'and new server process started')
    parser.add_argument('--peer', dest='peers', action='append', default=[],
                        help='peer location (in the form node:TCPport) to communicate')
    parser.add_argument('-d', '--debug', action='store_true', dest='loglevel', default=False,
                        help='if given, debug messages are printed')
    _discoro_config = vars(parser.parse_args(sys.argv[1:]))

    _discoro_var = _discoro_config.pop('config')
    if _discoro_var:
        import configparser
        cfg = configparser.ConfigParser()
        cfg.read(_discoro_var)
        cfg = dict(cfg.items('DEFAULT'))
        cfg['cpus'] = int(cfg['cpus'])
        cfg['udp_port'] = int(cfg['udp_port'])
        cfg['serve'] = int(cfg['serve'])
        cfg['msg_timeout'] = int(cfg['msg_timeout'])
        cfg['min_pulse_interval'] = int(cfg['min_pulse_interval'])
        cfg['max_pulse_interval'] = int(cfg['max_pulse_interval'])
        cfg['zombie_period'] = int(cfg['zombie_period'])
        cfg['ping_interval'] = int(cfg['ping_interval'])
        cfg['daemon'] = cfg['daemon'] == 'True'
        cfg['clean'] = cfg['clean'] == 'True'
        # cfg['discover_peers'] = cfg['discover_peers'] == 'True'
        cfg['loglevel'] = cfg['loglevel'] == 'True'
        cfg['tcp_ports'] = [_discoro_var.strip()[1:-1] for _discoro_var in
                            cfg['tcp_ports'][1:-1].split(',')]
        cfg['tcp_ports'] = [_discoro_var for _discoro_var in cfg['tcp_ports'] if _discoro_var]
        cfg['peers'] = [_discoro_var.strip()[1:-1] for _discoro_var in
                        cfg['peers'][1:-1].split(',')]
        cfg['peers'] = [_discoro_var for _discoro_var in cfg['peers'] if _discoro_var]
        for key, value in _discoro_config.items():
            if _discoro_config[key] != parser.get_default(key) or key not in cfg:
                cfg[key] = _discoro_config[key]
        _discoro_config = cfg
        del key, value, cfg
    del parser, MinPulseInterval, MaxPulseInterval
    del sys.modules['argparse'], globals()['argparse']

    _discoro_var = _discoro_config.pop('save_config')
    if _discoro_var:
        import configparser
        cfg = configparser.ConfigParser(_discoro_config)
        cfgfp = open(_discoro_var, 'w')
        cfg.write(cfgfp)
        cfgfp.close()
        exit(0)

    if not _discoro_config['min_pulse_interval']:
        _discoro_config['min_pulse_interval'] = MinPulseInterval
    if not _discoro_config['max_pulse_interval']:
        _discoro_config['max_pulse_interval'] = MaxPulseInterval
    if _discoro_config['msg_timeout'] < 1:
        raise Exception('msg_timeout must be at least 1')
    if (_discoro_config['min_pulse_interval'] and
        _discoro_config['min_pulse_interval'] < _discoro_config['msg_timeout']):
        raise Exception('min_pulse_interval must be at least msg_timeout')
    if (_discoro_config['max_pulse_interval'] and _discoro_config['min_pulse_interval'] and
       _discoro_config['max_pulse_interval'] < _discoro_config['min_pulse_interval']):
        raise Exception('max_pulse_interval must be at least min_pulse_interval')
    if _discoro_config['zombie_period']:
        if _discoro_config['zombie_period'] < _discoro_config['min_pulse_interval']:
            raise Exception('zombie_period must be at least min_pulse_interval')
    else:
        _discoro_config['zombie_period'] = 0

    _discoro_cpus = multiprocessing.cpu_count()
    if _discoro_config['cpus'] > 0:
        if _discoro_config['cpus'] > _discoro_cpus:
            raise Exception('CPU count must be <= %s' % _discoro_cpus)
        _discoro_cpus = _discoro_config['cpus']
    elif _discoro_config['cpus'] < 0:
        if -_discoro_config['cpus'] >= _discoro_cpus:
            raise Exception('CPU count must be > -%s' % _discoro_cpus)
        _discoro_cpus += _discoro_config['cpus']
    del _discoro_config['cpus']

    _discoro_tcp_ports = set()
    tcp_port = tcp_ports = None
    for tcp_port in _discoro_config.pop('tcp_ports', []):
        tcp_ports = tcp_port.split('-')
        if len(tcp_ports) == 1:
            _discoro_tcp_ports.add(int(tcp_ports[0]))
        elif len(tcp_ports) == 2:
            _discoro_tcp_ports = _discoro_tcp_ports.union(range(int(tcp_ports[0]),
                                                                int(tcp_ports[1]) + 1))
        else:
            raise Exception('Invalid TCP port range "%s"' % tcp_ports)
    _discoro_tcp_ports = sorted(_discoro_tcp_ports)

    if _discoro_tcp_ports:
        for tcp_port in range(_discoro_tcp_ports[-1] + 1,
                              _discoro_tcp_ports[-1] + 1 +
                              (_discoro_cpus + 1) - len(_discoro_tcp_ports)):
            _discoro_tcp_ports.append(int(tcp_port))
        # _discoro_tcp_ports = _discoro_tcp_ports[:(_discoro_cpus + 1)]
    else:
        _discoro_tcp_ports = [0] * (_discoro_cpus + 1)
    del tcp_port, tcp_ports

    peers, _discoro_config['peers'] = _discoro_config['peers'], []
    peer = None
    for peer in peers:
        peer = peer.split(':')
        if len(peer) != 2:
            raise Exception('peer "%s" is not valid' % ':'.join(peer))
        _discoro_config['peers'].append(asyncoro.serialize(asyncoro.Location(peer[0], peer[1])))
    del peer, peers

    _discoro_name = _discoro_config['name']
    if not _discoro_name:
        _discoro_name = socket.gethostname()
        if not _discoro_name:
            _discoro_name = 'discoro_server'

    _discoro_daemon = _discoro_config.pop('daemon', False)
    if not _discoro_daemon:
        try:
            if os.getpgrp() != os.tcgetpgrp(sys.stdin.fileno()):
                _discoro_daemon = True
        except:
            pass
        if os.name == 'nt':
            # Python 3 under Windows blocks multiprocessing.Process on reading
            # input; pressing "Enter" twice works (for one subprocess). Until
            # this is understood / fixed, disable reading input.
            print('\nReading standard input disabled, as multiprocessing does not seem to work'
                  'with reading input under Windows\n')
            _discoro_daemon = True

    _discoro_config['discover_peers'] = False

    # time at start of day
    _discoro_var = time.localtime()
    _discoro_var = (int(time.time()) - (_discoro_var.tm_hour * 3600) -
                    (_discoro_var.tm_min * 60))
    _discoro_service_start = _discoro_service_stop = _discoro_service_end = None
    if _discoro_config['service_start']:
        _discoro_service_start = time.strptime(_discoro_config.pop('service_start'), '%H:%M')
        _discoro_service_start = (_discoro_var + (_discoro_service_start.tm_hour * 3600) +
                                  (_discoro_service_start.tm_min * 60))
    if _discoro_config['service_stop']:
        _discoro_service_stop = time.strptime(_discoro_config.pop('service_stop'), '%H:%M')
        _discoro_service_stop = (_discoro_var + (_discoro_service_stop.tm_hour * 3600) +
                                 (_discoro_service_stop.tm_min * 60))
    if _discoro_config['service_end']:
        _discoro_service_end = time.strptime(_discoro_config.pop('service_end'), '%H:%M')
        _discoro_service_end = (_discoro_var + (_discoro_service_end.tm_hour * 3600) +
                                (_discoro_service_end.tm_min * 60))

    if (_discoro_service_start or _discoro_service_stop or _discoro_service_end):
        if not _discoro_service_start:
            _discoro_service_start = int(time.time())
        if _discoro_service_stop:
            if _discoro_service_start >= _discoro_service_stop:
                raise Exception('"service_start" must be before "service_stop"')
        if _discoro_service_end:
            if _discoro_service_start >= _discoro_service_end:
                raise Exception('"service_start" must be before "service_end"')
            if _discoro_service_stop and _discoro_service_stop >= _discoro_service_end:
                raise Exception('"service_stop" must be before "service_end"')
        if not _discoro_service_stop and not _discoro_service_end:
            raise Exception('"service_stop" or "service_end" must also be given')

    if _discoro_config['max_file_size']:
        _discoro_var = re.match(r'(\d+)([kKmMgGtT]?)', _discoro_config['max_file_size'])
        if not _discoro_var or len(_discoro_var.group(0)) != len(_discoro_config['max_file_size']):
            raise Exception('Invalid max_file_size option')
        _discoro_config['max_file_size'] = int(_discoro_var.group(1))
        if _discoro_var.group(2):
            _discoro_var = _discoro_var.group(2).lower()
            _discoro_config['max_file_size'] *= 1024**({'k': 1, 'm': 2, 'g': 3,
                                                        't': 4}[_discoro_var])
    else:
        _discoro_config['max_file_size'] = 0

    if _discoro_config['certfile']:
        _discoro_config['certfile'] = os.path.abspath(_discoro_config['certfile'])
    else:
        _discoro_config['certfile'] = None
    if _discoro_config['keyfile']:
        _discoro_config['keyfile'] = os.path.abspath(_discoro_config['keyfile'])
    else:
        _discoro_config['keyfile'] = None

    _discoro_node_auth = ''.join(hex(_)[2:] for _ in os.urandom(10))
    _discoro_node_auth = hashlib.sha1(_discoro_node_auth.encode()).hexdigest()

    class _discoro_Struct(object):

        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

        def __setattr__(self, name, value):
            if hasattr(self, name):
                self.__dict__[name] = value
            else:
                raise AttributeError('Invalid attribute "%s"' % name)

    _discoro_spawn_proc = None
    _discoro_busy_time = multiprocessing.Value('I', 0)
    _discoro_mp_queue = multiprocessing.Queue()
    _discoro_servers = [None] * _discoro_cpus
    for _discoro_server_id in range(1, _discoro_cpus + 1):
        _discoro_server = _discoro_Struct(id=_discoro_server_id, proc=None, coro=None,
                                          name='%s_proc-%s' % (_discoro_name, _discoro_server_id))
        _discoro_servers[_discoro_server_id - 1] = _discoro_server

    def _discoro_node_proc(coro=None):
        from asyncoro.discoro import DiscoroNodeAvailInfo, DiscoroNodeInfo, MaxPulseInterval
        global _discoro_servers, _discoro_config, _discoro_spawn_proc

        coro.register('discoro_node')
        coro_scheduler = asyncoro.AsynCoro.instance()
        last_pulse = last_ping = time.time()
        scheduler_coro = cur_computation_auth = None
        interval = _discoro_config['max_pulse_interval']
        ping_interval = _discoro_config.pop('ping_interval')
        msg_timeout = _discoro_config['msg_timeout']
        zombie_period = _discoro_config['zombie_period']
        disk_path = coro_scheduler.dest_path
        _discoro_config['node_coro'] = asyncoro.serialize(coro)
        _discoro_config['node'] = coro.location.addr

        def monitor_peers(coro=None):
            coro.set_daemon()
            while 1:
                msg = yield coro.receive()
                if not isinstance(msg, asyncoro.PeerStatus):
                    continue
                if msg.status == asyncoro.PeerStatus.Offline:
                    if (scheduler_coro and scheduler_coro.location == msg.location):
                        _discoro_node_coro.send({'req': 'release', 'auth': cur_computation_auth})

        def mp_queue_server():
            global _discoro_tcp_ports
            while 1:
                proc_id, proc_coro = _discoro_mp_queue.get(block=True)
                server = _discoro_servers[proc_id - 1]
                if proc_coro:
                    server.coro = asyncoro.deserialize(proc_coro)
                    # if not _discoro_tcp_ports[server.id - 1]:
                    #     _discoro_tcp_ports[server.id - 1] = server.coro.location.port
                else:
                    server.coro = None
                    if _discoro_config['serve']:
                        if scheduler_coro and service_available(now):
                            asyncoro.logger.warning('Server %s terminated', server.name)
                            # _discoro_start_server(server)
                    elif all(not server.coro for server in _discoro_servers):
                        _discoro_node_coro.send({'req': 'quit', 'auth': _discoro_node_auth})
                        break

        def service_available(now):
            if not _discoro_config['serve']:
                return False
            if not _discoro_service_start:
                return True
            if _discoro_service_stop:
                if (_discoro_service_start <= now < _discoro_service_stop):
                    return True
            else:
                if (_discoro_service_start <= now < _discoro_service_end):
                    return True
            return False

        def service_times_proc(coro=None):
            global _discoro_service_start, _discoro_service_stop, _discoro_service_end
            coro.set_daemon()
            while 1:
                if _discoro_service_stop:
                    now = int(time.time())
                    yield coro.sleep(_discoro_service_stop - now)
                    for server in _discoro_servers:
                        if server.coro:
                            server.coro.send({'req': 'quit', 'node_auth': _discoro_node_auth})

                if _discoro_service_end:
                    now = int(time.time())
                    yield coro.sleep(_discoro_service_end - now)
                    for server in _discoro_servers:
                        if server.coro:
                            server.coro.send({'req': 'terminate', 'node_auth': _discoro_node_auth})

                # advance times for next day
                _discoro_service_start += 24 * 3600
                if _discoro_service_stop:
                    _discoro_service_stop += 24 * 3600
                if _discoro_service_end:
                    _discoro_service_end += 24 * 3600
                # disable service till next start
                coro_scheduler.ignore_peers(True)
                now = int(time.time())
                yield coro.sleep(_discoro_service_start - now)
                coro_scheduler.ignore_peers(False)
                coro_scheduler.discover_peers(port=_discoro_config['scheduler_port'])

        def close_computation():
            global _discoro_spawn_proc

            for server in _discoro_servers:
                if server.coro:
                    server.coro.send({'req': 'quit', 'node_auth': _discoro_node_auth})
            if _discoro_spawn_proc:
                _discoro_send_pipe.send({'req': 'quit'})
                if _discoro_send_pipe.poll(5) and _discoro_send_pipe.recv() == 'closed':
                    _discoro_spawn_proc = None
                else:
                    _discoro_spawn_proc.terminate()
                    _discoro_spawn_proc = None
                while _discoro_send_pipe.poll():  # clear pipe
                    _discoro_send_pipe.recv()
                while _discoro_recv_pipe.poll():  # clear pipe
                    _discoro_recv_pipe.recv()
            for name in os.listdir(_discoro_config['dest_path']):
                if name.startswith('discoroproc-') or name == 'discoroscheduler':
                    continue
                name = os.path.join(_discoro_config['dest_path'], name)
                if os.path.isdir(name):
                    shutil.rmtree(name, ignore_errors=True)
                elif os.path.isfile(name):
                    try:
                        os.remove(name)
                    except:
                        pass
            coro_scheduler.discover_peers(port=_discoro_config['scheduler_port'])

        if _discoro_service_start:
            asyncoro.Coro(service_times_proc)

        qserver = threading.Thread(target=mp_queue_server)
        qserver.daemon = True
        qserver.start()
        coro_scheduler.peer_status(asyncoro.Coro(monitor_peers))
        coro_scheduler.discover_peers(port=_discoro_config['scheduler_port'])
        for peer in _discoro_config['peers']:
            asyncoro.Coro(coro_scheduler.peer, asyncoro.deserialize(peer))

        # TODO: create new pipe for each computation instead?
        _discoro_recv_pipe, _discoro_send_pipe = multiprocessing.Pipe(duplex=True)

        while 1:
            msg = yield coro.receive(timeout=interval)
            now = time.time()
            if msg:
                try:
                    req = msg['req']
                except:
                    req = ''

                if req == 'server_setup':
                    try:
                        server = _discoro_servers[msg['id'] - 1]
                        assert msg['auth'] == cur_computation_auth
                    except:
                        pass
                    else:
                        if not server.coro:
                            server.coro = msg['coro']
                        last_pulse = now

                elif req == 'discoro_node_info':
                    # request from scheduler
                    client = msg.get('client', None)
                    if isinstance(client, asyncoro.Coro):
                        if psutil:
                            info = DiscoroNodeAvailInfo(coro.location, 100.0 - psutil.cpu_percent(),
                                                        psutil.virtual_memory().available,
                                                        psutil.disk_usage(disk_path).free,
                                                        100.0 - psutil.swap_memory().percent)
                        else:
                            info = DiscoroNodeAvailInfo(coro.location, None, None, None, None)
                        info = DiscoroNodeInfo(_discoro_name, coro.location.addr,
                                               len(_discoro_servers), platform.platform(), info)
                        client.send(info)

                elif req == 'reserve':
                    # request from scheduler
                    client = msg.get('client', None)
                    cpus = msg.get('cpus', -1)
                    auth = msg.get('auth', None)
                    if (isinstance(client, asyncoro.Coro) and isinstance(cpus, int) and
                        cpus >= 0 and not cur_computation_auth and not scheduler_coro and
                        service_available(now) and (len(_discoro_servers) >= cpus) and auth and
                        isinstance(msg.get('status_coro', None), asyncoro.Coro) and
                        isinstance(msg.get('computation_location', None), asyncoro.Location)):
                        if (yield coro_scheduler.peer(msg['computation_location'])):
                            cpus = 0
                        else:
                            close_computation()
                            for server in _discoro_servers:
                                if server.coro:
                                    yield coro.sleep(0.1)
                            if not cpus:
                                cpus = len(_discoro_servers)
                        if ((yield client.deliver(cpus, timeout=min(msg_timeout, interval))) == 1 and
                            cpus):
                            cur_computation_auth = auth
                            last_pulse = now
                            _discoro_busy_time.value = int(time.time())
                            scheduler_coro = msg['status_coro']
                    elif isinstance(client, asyncoro.Coro):
                        client.send(0)

                elif req == 'computation':
                    client = msg.get('client', None)
                    computation = msg.get('computation', None)
                    if (cur_computation_auth == msg.get('auth', None) and
                        isinstance(client, asyncoro.Coro) and isinstance(computation, Computation)):
                        interval = computation._pulse_interval
                        last_pulse = now
                        if interval:
                            interval = min(interval, _discoro_config['max_pulse_interval'])
                        else:
                            interval = _discoro_config['max_pulse_interval']
                        _discoro_busy_time.value = int(time.time())
                        _discoro_config['scheduler_coro'] = asyncoro.serialize(scheduler_coro)
                        _discoro_config['computation_auth'] = computation._auth
                        id_ports = [(server.id, _discoro_tcp_ports[server.id - 1])
                                    for server in _discoro_servers if not server.coro]
                        args = (_discoro_config, id_ports, _discoro_mp_queue, _discoro_recv_pipe,
                                asyncoro.serialize(computation) if os.name == 'nt'
                                else computation, msg.get('setup_args', ()))
                        _discoro_spawn_proc = multiprocessing.Process(target=_discoro_spawn,
                                                                      args=args)
                        _discoro_spawn_proc.start()
                        if _discoro_send_pipe.poll(10):
                            cpus = _discoro_send_pipe.recv()
                        else:
                            cpus = 0
                        if ((yield client.deliver(cpus)) == 1) and cpus:
                            pass
                        else:
                            close_computation()

                elif req == 'release':
                    auth = msg.get('auth', None)
                    if cur_computation_auth and auth == cur_computation_auth:
                        close_computation()
                        cur_computation_auth = None
                        scheduler_coro = None
                        interval = MaxPulseInterval
                        released = 'released'
                    else:
                        released = 'invalid'
                    client = msg.get('client', None)
                    if isinstance(client, asyncoro.Coro):
                        client.send(released)
                    if released == 'released' and _discoro_config['serve'] > 0:
                        _discoro_config['serve'] -= 1
                        if not _discoro_config['serve']:
                            break

                elif req == 'close' or req == 'quit' or req == 'terminate':
                    auth = msg.get('auth', None)
                    if auth == _discoro_node_auth:
                        close_computation()
                        cur_computation_auth = None
                        scheduler_coro = None
                        interval = MaxPulseInterval
                        if req == 'quit' or req == 'terminate':
                            _discoro_config['serve'] = 0
                            if all(not server.coro for server in _discoro_servers):
                                # _discoro_mp_queue.close()
                                _discoro_send_pipe.close()
                                _discoro_recv_pipe.close()
                                break

                else:
                    asyncoro.logger.warning('Invalid message %s ignored',
                                            str(msg) if isinstance(msg, dict) else '')

            if scheduler_coro:
                scoro = scheduler_coro  # copy in case scheduler closes meanwhile
                msg = {'status': 'pulse', 'location': coro.location}
                if psutil:
                    msg['node_status'] = DiscoroNodeAvailInfo(
                        coro.location, 100.0 - psutil.cpu_percent(),
                        psutil.virtual_memory().available, psutil.disk_usage(disk_path).free,
                        100.0 - psutil.swap_memory().percent)

                sent = yield scoro.deliver(msg, timeout=msg_timeout)
                if sent == 1:
                    last_pulse = now
                elif (now - last_pulse) > (5 * interval):
                    asyncoro.logger.warning('Scheduler is not reachable; closing computation "%s"',
                                            cur_computation_auth)
                    close_computation()
                    cur_computation_auth = None
                    scheduler_coro = None
                    interval = MaxPulseInterval
                    asyncoro.Coro(coro_scheduler.close_peer, scoro.location)
                    if _discoro_config['serve'] > 0:
                        _discoro_config['serve'] -= 1
                        if not _discoro_config['serve']:
                            break

                if (zombie_period and ((now - _discoro_busy_time.value) > zombie_period) and
                    cur_computation_auth):
                    asyncoro.logger.warning('Closing zombie computation "%s"', cur_computation_auth)
                    close_computation()
                    cur_computation_auth = None
                    scheduler_coro = None
                    interval = MaxPulseInterval
                    if _discoro_config['serve'] > 0:
                        _discoro_config['serve'] -= 1
                        if not _discoro_config['serve']:
                            break

            if ping_interval and (now - last_ping) > ping_interval and service_available(now):
                coro_scheduler.discover_peers(port=_discoro_config['scheduler_port'])

        try:
            os.remove(_discoro_node_pid_file)
        except:
            pass
        os.kill(os.getpid(), signal.SIGINT)

    _discoro_server_config = {}
    for _discoro_var in ['udp_port', 'tcp_port', 'node', 'ext_ip_addr', 'name',
                         'discover_peers', 'secret', 'certfile', 'keyfile', 'dest_path',
                         'max_file_size']:
        _discoro_server_config[_discoro_var] = _discoro_config.get(_discoro_var, None)
    _discoro_server_config['name'] = '%s_proc-0' % _discoro_name
    _discoro_server_config['tcp_port'] = _discoro_tcp_ports.pop(0)
    if _discoro_config['loglevel']:
        asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
        # asyncoro.logger.show_ms(True)
    else:
        asyncoro.logger.setLevel(asyncoro.Logger.INFO)
    _discoro_scheduler = asyncoro.AsynCoro(**_discoro_server_config)
    _discoro_scheduler.dest_path = os.path.join(_discoro_scheduler.dest_path, 'discoro')
    _discoro_node_pid_file = os.path.join(_discoro_scheduler.dest_path, 'discoroproc-0.pid')
    if _discoro_config['clean']:
        try:
            os.remove(_discoro_node_pid_file)
        except:
            pass
    try:
        _discoro_var = os.open(_discoro_node_pid_file, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o600)
        os.write(_discoro_var, str(os.getpid()).encode())
        os.close(_discoro_var)
    except:
        raise Exception('Another discoronode seem to be running; '
                        'check no discoronode and servers are running and '
                        'remove *.pid files in %s' % _discoro_scheduler.dest_path)

    _discoro_config['name'] = _discoro_name
    _discoro_config['dest_path'] = _discoro_scheduler.dest_path
    _discoro_config['node_auth'] = _discoro_node_auth
    _discoro_config['busy_time'] = _discoro_busy_time
    _discoro_node_coro = asyncoro.Coro(_discoro_node_proc)
    del _discoro_server_config, _discoro_var

    def sighandler(signum, frame):
        if os.path.isfile(_discoro_node_pid_file):
            _discoro_node_coro.send({'req': 'quit', 'auth': _discoro_node_auth})
        else:
            raise KeyboardInterrupt

    try:
        signal.signal(signal.SIGHUP, sighandler)
        signal.signal(signal.SIGQUIT, sighandler)
    except:
        pass
    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGABRT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)
    del sighandler

    if _discoro_daemon:
        while 1:
            try:
                time.sleep(3600)
            except:
                if os.path.exists(_discoro_node_pid_file):
                    _discoro_node_coro.send({'req': 'quit', 'auth': _discoro_node_auth})
                break
    else:
        while 1:
            # wait a bit for any output for previous command is done
            time.sleep(0.2)
            try:
                _discoro_cmd = input(
                        '\nEnter\n'
                        '  "status" to get status\n'
                        '  "close" to stop accepting new jobs and\n'
                        '          close current computation when current jobs are finished\n'
                        '  "quit" to "close" current computation and exit discoronode\n'
                        '  "terminate" to kill current jobs and "quit": ')
            except:
                if os.path.exists(_discoro_node_pid_file):
                    _discoro_node_coro.send({'req': 'quit', 'auth': _discoro_node_auth})
                break
            else:
                _discoro_cmd = _discoro_cmd.strip().lower()
                if not _discoro_cmd:
                    _discoro_cmd = 'status'

            print('')
            if _discoro_cmd == 'status':
                for _discoro_server in _discoro_servers:
                    if _discoro_server.coro:
                        _discoro_server.coro.send({'req': _discoro_cmd,
                                                   'node_auth': _discoro_node_auth})
                    else:
                        print('  discoro server "%s" is not currently used' %
                              _discoro_server.name)
            elif _discoro_cmd in ('close', 'quit', 'terminate'):
                _discoro_node_coro.send({'req': _discoro_cmd, 'auth': _discoro_node_auth})
                break

    try:
        _discoro_node_coro.value()
    except:
        pass
    exit(0)



# Run 'discoronode.py' program to start processes to execute computations sent
# by this client, along with this program.

# This example shows how to use message passing, transferring files.
# The input files are assumed to be in current working directory as
# 'input0.dat', 'input1.dat', ..., 'input9.dat' (10 input files).
# 'run_jobs_proc' creates jobs as and when a discoro process is found,
# or a process finishes a job. Each job is processed by a remote coroutine
# (rcoro) and a corresponding local coroutine (with 'client_proc')
# interacts with that rcoro to send the input data to where rcoro is,
# an instance of class C to give the information needed to process the
# file. Client then waits for rcoro to finish processing job, send the resulting
# output file in the form 'outputX.dat' and result (in this case, it is
# the return value of 'send_file').

# Note that the objective is to illustrate features, so implementation
# is not ideal. Error checking is skipped at a few places for brevity.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *

# objects of C are sent by a client to remote coroutine
class C(object):
    def __init__(self, i, data_file, n, client):
        self.i = i
        self.data_file = data_file
        self.result_file = None
        self.n = n
        self.client = client


# this generator function is sent to remote server to run
# coroutines there
def rcoro_proc(coro=None):
    import os
    # receive object from client_proc coroutine
    cobj = yield coro.receive()
    if not cobj:
        raise StopIteration
    # Input file is already copied at where this rcoro is running (by client).
    # For given input file, create an output file with each line in the output
    # file computed as length of corresponding line in input file
    cobj.result_file = 'result-%s' % cobj.data_file
    with open(cobj.data_file, 'r') as data_fd:
        with open(cobj.result_file, 'w') as result_fd:
            for lineno, line in enumerate(data_fd, start=1):
                result_fd.write('%d: %d\n' % (lineno, len(line)-1))
    # 'sleep' to simulate computing
    yield coro.sleep(cobj.n)
    # transfer the result file to client
    status = yield asyncoro.AsynCoro().send_file(cobj.client.location, cobj.result_file,
                                                 overwrite=True, timeout=30)
    if status:
        print('Could not send %s to %s' % (cobj.result_file, cobj.client.location))
        cobj.result_file = None
    cobj.client.send(cobj)
    os.remove(cobj.data_file)
    os.remove(cobj.result_file)


# this generator function is used to create local coroutine (at the
# client) to communicate with a remote coroutine
def client_proc(job_id, data_file, rcoro, coro=None):
    # send input file to rcoro.location; this will be saved to discoro process's
    # working directory
    if (yield asyncoro.AsynCoro().send_file(rcoro.location, data_file, timeout=10)) < 0:
        print('Could not send input data to %s' % rcoro.location)
        # terminate remote coro
        rcoro.send(None)
        raise StopIteration(-1)
    # send info about input
    obj = C(job_id, data_file, random.uniform(5, 8), coro)
    if (yield rcoro.deliver(obj)) != 1:
        print('Could not send input to %s' % rcoro.location)
        raise StopIteration(-1)
    # rcoro sends result to this coroutine as message
    result = yield coro.receive()
    if not result.result_file:
        print('Processing %s failed' % obj.i)
        raise StopIteration(-1)
    # rcoro saves results file at this client, which is saved in asyncoro's
    # dest_path, not current working directory!
    result_file = os.path.join(asyncoro.AsynCoro().dest_path, result.result_file)
    # move file to cwd
    target = os.path.join(os.getcwd(), os.path.basename(result_file))
    os.rename(result_file, target)
    print('    job %s output is in %s' % (obj.i, target))


def run_jobs_proc(computation, data_files, coro=None):
    # schedule computation with the scheduler; scheduler accepts one computation
    # at a time, so if scheduler is shared, the computation is queued until it
    # is done with already scheduled computations
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    for i in range(len(data_files)):
        data_file = data_files[i]
        # create remote coroutine
        rcoro = yield computation.run(rcoro_proc)
        if isinstance(rcoro, asyncoro.Coro):
            # create local coroutine to send input file and data to rcoro
            asyncoro.Coro(client_proc, i, data_file, rcoro)
        else:
            print('  job %s failed: %s' % (i, rcoro))

    yield computation.close()


if __name__ == '__main__':
    import random, os, sys, glob
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    if os.path.dirname(sys.argv[0]):
        os.chdir(os.path.dirname(sys.argv[0]))
    data_files = glob.glob('discoro_client*.py')
    if not data_files:
        raise Exception('No data files to process')
    if len(sys.argv) > 1:
        data_files = data_files[:int(sys.argv[1])]

    # if scheduler is not already running (on a node as a program),
    # start it (private scheduler):
    Scheduler()
    # unlike in earlier examples, rcoro_proc is not sent with
    # computation (as it is not included in 'components';
    # instead, it is sent each time a job is submitted,
    # which is a bit inefficient
    computation = Computation([C])

    asyncoro.Coro(run_jobs_proc, computation, data_files)



# Asynchronous file example. Works with Linux, OS X and other Unix
# variants, but not Windows, as in Windows sockets don't support file
# I/O for asynchronous I/O.

# The file descriptor is associated with socket. The client sends data
# in chunks and server reads lines from the data receivd from
# client. Both compute checksum to check that data is received
# correctly.

# argv[1] must be a text file

import socket, hashlib, sys, os
import asyncoro
import asyncoro.asyncfile

def client_proc(host, port, input, coro=None):
    # client reads input file and sends data in chunks
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock = asyncoro.AsyncSocket(sock)
    yield sock.connect((host, port))
    # data can be written to this asynchronous socket; however, for
    # illustration, convert its file descriptor to asynchronous file
    # and write to that instead
    afd = asyncoro.asyncfile.AsyncFile(sock)
    input = open(input)
    csum = hashlib.sha1()
    while True:
        data = os.read(input.fileno(), 16*1024)
        if not data:
            break
        csum.update(data)
        n = yield afd.write(data, full=True)
    afd.close()
    print('client sha1 csum: %s' % csum.hexdigest())

def server_proc(conn, coro=None):
    # conn is a synchronous socket (as it is obtained from synchronous
    # 'accept'); it's file-descriptor is converted to asynchronous
    # file to read data from that
    afd = asyncoro.asyncfile.AsyncFile(conn)
    csum = hashlib.sha1()
    nlines = 0
    while True:
        # read lines from data
        line = yield afd.readline()
        if not line:
            break
        csum.update(line)
        nlines += 1
    afd.close()
    print('server sha1 csum: %s' % (csum.hexdigest()))
    print('lines: %s' % (nlines))

asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
sock.bind(('localhost', 0))
sock.listen(5)
host, port = sock.getsockname()
print('host: %s, port: %s' % (host, port))

asyncoro.Coro(client_proc, host, port, sys.argv[1] if len(sys.argv) > 1 else sys.argv[0])

conn, addr = sock.accept()
asyncoro.Coro(server_proc, conn)



# Run 'discoronode.py' program to start processes to execute computations sent
# by this client, along with this program.

# This example illustrates in-memory processing with 'node_setup' to read date
# in to memory. Remote coroutines ('compute' in this case) then process data in
# memory. This example works with POSIX (Linux, OS X etc.), but not Windows. See
# 'discoro_client9_server.py' which initializes each server to read data in to
# memory for processing in computations.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *

def node_available(avail_info, data_file, coro=None):
    import os
    # 'node_available' is executed locally (at client) when a node is
    # available. 'location' is Location instance of node. When this coroutine
    # is executed, 'depends' of computation would've been transferred.

    # data_file could've been sent with the computation 'depends'; however, to
    # illustrate how files can be sent separately (e.g., to transfer different
    # files to different nodes), file is transferred with
    # 'node_available'.

    print('  Sending %s to %s' % (data_file, avail_info.location.addr))
    sent = yield asyncoro.AsynCoro().send_file(avail_info.location, data_file,
                                               overwrite=True, timeout=5)
    if (sent < 0):
        print('Could not send data file "%s" to %s' % (data_file, avail_info.location))
        raise StopIteration(-1)

    # value of coroutine (last value yield'ed or value of 'raise StopIteration')
    # will be passed to node_setup as argument(s).
    yield computation.enable_node(avail_info.location.addr, data_file)

def node_setup(data_file):
    # 'node_setup' is executed on a node with the arguments returned by
    # 'node_available'. This coroutine should return 0 to indicate successful
    # initialization.

    # variables declared as 'global' will be available (as read-only) in
    # coroutines.
    global os, hashlib, data, file_name
    import os, hashlib
    # note that files transferred to node are in parent directory of cwd where
    # each computation is run (in case such files need to be accessed in
    # computation).
    print('data_file: "%s"' % data_file)
    with open(data_file, 'rb') as fd:
        data = fd.read()
    os.remove(data_file)  # data_file is not needed anymore
    file_name = data_file
    yield 0  # coroutine must have at least one 'yield' and 0 indicates success

# 'compute' is executed at remote server process repeatedly to compute checksum
# of data in memory, initialized by 'node_setup'
def compute(alg, n, coro=None):
    global data, hashlib
    yield coro.sleep(n)
    checksum = getattr(hashlib, alg)()
    checksum.update(data)
    raise StopIteration((file_name, alg, checksum.hexdigest()))

# local coroutine to process status messages from scheduler
def status_proc(coro=None):
    coro.set_daemon()
    i = 0
    while 1:
        msg = yield coro.receive()
        if not isinstance(msg, DiscoroStatus):
            continue
        if msg.status == Scheduler.NodeDiscovered:
            asyncoro.Coro(node_available, msg.info.avail_info, data_files[i])
            i += 1

def client_proc(computation, coro=None):
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    # execute 10 jobs (coroutines) and get their results. Note that number of
    # jobs created can be more than number of server processes available; the
    # scheduler will use as many processes as necessary/available, running one
    # job at a server process
    yield coro.sleep(2)
    algorithms = ['md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512']
    args = [(algorithms[i % len(algorithms)], random.uniform(1, 3)) for i in range(15)]
    results = yield computation.run_results(compute, args)
    for i, result in enumerate(results):
        if isinstance(result, tuple) and len(result) == 3:
            print('   %ssum for %s: %s' % (result[1], result[0], result[2]))
        else:
            print('  rcoro failed for %s: %s' % (args[i][0], str(result)))

    yield computation.close()


if __name__ == '__main__':
    import logging, random, functools, sys, os, glob
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    if os.path.dirname(sys.argv[0]):
        os.chdir(os.path.dirname(sys.argv[0]))
    # if scheduler is not already running (on a node as a program),
    # start private scheduler:
    Scheduler()

    data_files = glob.glob('discoro_client*.py')

    # Since this example doesn't work with Windows, 'node_allocations' feature
    # is used to filter out nodes running Windows.
    node_allocations = [DiscoroNodeAllocate(node='*', platform='Windows', cpus=0)]
    computation = Computation([compute], node_allocations=node_allocations,
                              status_coro=asyncoro.Coro(status_proc), node_setup=node_setup,
                              disable_nodes=True,
                              )
    asyncoro.Coro(client_proc, computation)



# server program for processing requests received with message passing
# (asynchronous concurrent programming) from remote client
# (tut_client.py) on same network;
# see http://asyncoro.sourceforge.net/tutorial.html for details.

import sys
import asyncoro.disasyncoro as asyncoro

def server_proc(coro=None):
    coro.set_daemon()
    coro.register('server_coro')
    while True:
        msg = yield coro.receive()
        print('processing %s' % (msg))

asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
scheduler = asyncoro.AsynCoro(udp_port=0)
server = asyncoro.Coro(server_proc)
if sys.version_info.major > 2:
    read_input = input
else:
    read_input = raw_input
while True:
    try:
        cmd = read_input().strip().lower()
        if cmd in ('quit', 'exit'):
            break
    except:
        break



# Run 'discoronode.py' program on Amazon EC2 cloud computing and run this
# program on local computer.

# in this example, ssh is used for port forwarding make sure EC2 instance allows
# inbound TCP port 51347 (and any additional ports, depending on how many CPUs
# are used by discoronode) assume '54.204.242.185' is external IP address of EC2
# instance; login to that node with the PEM key: 'ssh -i my-key.pem
# 4567:127.0.0.1:4567 54.204.242.185' run discoronode on EC2 at port (starting
# with) 51347 with: 'discoronode.py -d --ext_ip_addr 54.204.242.185 --tcp_ports
# 51347'

# Distributed computing example where this client sends computation to remote
# discoro process to run as remote coroutines. At any time at most one
# computation coroutine is scheduled at a process (due to
# RemoteCoroScheduler). This example shows how to use 'execute' method of
# RemoteCoroScheduler to run comutations and get their results easily.

# This example can be combined with in-memory processing (see
# 'discoro_client9_node.py') and streaming (see 'discoro_client6.py') for
# efficient processing of data and communication.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *


# this generator function is sent to remote discoro servers to run
# coroutines there
def compute(n, coro=None):
    import time
    yield coro.sleep(n)
    raise StopIteration(time.asctime()) # result of 'compute' is current time


def client_proc(computation, njobs, coro=None):
    # schedule computation with the scheduler; scheduler accepts one computation
    # at a time, so if scheduler is shared, the computation is queued until it
    # is done with already scheduled computations
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    # pair EC2 node with this client with:
    yield asyncoro.AsynCoro().peer(asyncoro.Location('54.204.242.185', 51347))
    # if multiple nodes are used, 'broadcast' option can be used to pair with
    # all nodes with just one statement as:
    # yield asyncoro.AsynCoro().peer(asyncoro.Location('54.204.242.185', 51347), broadcast=True)

    # execute n jobs (coroutines) and get their results. Note that number of
    # jobs created can be more than number of server processes available; the
    # scheduler will use as many processes as necessary/available, running one
    # job at a server process
    args = [random.uniform(3, 10) for _ in range(njobs)]
    results = yield computation.run_results(compute, args)
    for result in results:
        print('result: %s' % result)

    yield computation.close()


if __name__ == '__main__':
    import sys, random
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    asyncoro.AsynCoro(node='127.0.0.1', tcp_port=4567)
    njobs = 10 if len(sys.argv) == 1 else int(sys.argv[1])
    # if scheduler is not already running (on a node as a program),
    # start private scheduler:
    Scheduler()
    # send 'compute' generator function
    computation = Computation([compute])
    asyncoro.Coro(client_proc, computation, njobs)



# Run 'discoronode.py' program to start processes to execute computations sent
# by this client, along with this program.

# Example where this client sends computation to remote discoro process to run
# as remote coroutines. Remote coroutines and client use message passing to
# exchange objects (instance of class 'C'). Instead of using 'map_results' as
# done in 'discoro_client1.py', remote coroutine sends the result back to client
# with message passing.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *


# objects of C are exchanged between client and servers
class C(object):
    def __init__(self, i):
        self.i = i
        self.n = None

    def __repr__(self):
        return '%d: %s' % (self.i, self.n)


# this generator function is sent to remote discoro servers to run coroutines
# there
def compute(obj, client, coro=None):
    # obj is an instance of C
    import math
    # this coroutine and client can use message passing
    print('process at %s received: %s' % (coro.location, obj.n))
    yield coro.sleep(obj.n)
    obj.n = math.sqrt(obj.n)
    # send result back to client
    yield client.deliver(obj, timeout=5)


def client_proc(computation, njobs, coro=None):
    # schedule computation with the scheduler; scheduler accepts one computation
    # at a time, so if scheduler is shared, the computation is queued until it
    # is done with already scheduled computations
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    # create a separate coroutine to receive results, so they can be processed
    # as soon as received
    def recv_results(coro=None):
        for i in range(njobs):
            msg = yield coro.receive()
            print('    result for job %d: %s' % (i, msg))

    # remote coroutines send replies as messages to this coro
    results_coro = asyncoro.Coro(recv_results)

    # run njobs; each job will be executed by one discoro server
    for i in range(njobs):
        cobj = C(i)
        cobj.n = random.uniform(5, 10)
        # as noted in 'discoro_client2.py', 'run' method is used to run jobs
        # sequentially; use 'run_async' to run multiple jobs on one server
        # concurrently
        print('  request %d: %s' % (i, cobj.n))
        rcoro = yield computation.run(compute, cobj, results_coro)
        if not isinstance(rcoro, asyncoro.Coro):
            print('failed to create rcoro %s: %s' % (i, rcoro))

    # wait for all results and close computation
    yield computation.close()


if __name__ == '__main__':
    import random, sys
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is not already running (on a node as a program),
    # start it (private scheduler):
    Scheduler()
    # send generator function and class C (as the computation uses
    # objects of C)
    computation = Computation([compute, C])
    # run 10 (or given number of) jobs
    asyncoro.Coro(client_proc, computation, 10 if len(sys.argv) < 2 else int(sys.argv[1])).value()



#!/usr/bin/env python

# example where client and server communicate through a channel.
# use with its server 'remote_channel_server.py'

import sys
# import disasyncoro to use distributed version of AsynCoro
import asyncoro.disasyncoro as asyncoro

def sender_proc(rchannel, coro=None):
    # send messages to channel; 'deliver' is used with n=2, so
    # messages will not be sent until both receivers subscribe to
    # channel
    for x in range(10):
        msg = {'msg':'message %s' % x, 'sender':coro}
        n = yield rchannel.deliver(msg, n=2)
        print('delivered to: %s' % n)
    rchannel.send(None)

def receiver_proc2(coro=None):
    # if server is in remote network, add it explicitly
    # yield scheduler.peer('remote.peer.ip', stream_send=True)
    # get reference to remote channel in server
    rchannel = yield asyncoro.Channel.locate('2clients')
    # this coro subscribes to the channel, so gets messages to server
    # channel
    print('server is at %s' % rchannel.location)
    if (yield rchannel.subscribe(coro)) != 0:
        raise Exception('subscription failed')
    asyncoro.Coro(sender_proc, rchannel)
    while True:
        msg = yield coro.receive()
        if msg is None:
            break
        print('Received "%s" from %s at %s' % \
              (msg['msg'], msg['sender'].name, msg['sender'].location))
    yield rchannel.unsubscribe(coro)

asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
# scheduler = asyncoro.AsynCoro()
asyncoro.Coro(receiver_proc2)



#!/usr/bin/env python

# chat server; must be used with 'chat_sock_client.py'

import sys, socket, collections
# messages are sent as strings, instead of bytes, so this program
# needs to be modified to work with Python 3
import asyncoro

def client_proc(conn, addr, server, coro=None):
    # given conn is synchronous, convert it to asynchronous
    conn = asyncoro.AsyncSocket(conn)
    server.send(('joined'.encode(), (conn, addr)))

    while True:
        line = yield conn.recv_msg()
        if not line:
            server.send(('left'.encode(), (conn, addr)))
            break
        server.send(('broadcast'.encode(), (conn, line)))

def server_proc(coro=None):
    clients = set()
    while True:
        cmd, item = yield coro.receive()
        cmd = cmd.decode()
        if cmd == 'broadcast':
            conn, msg = item
            msg = ('%s says: %s' % (conn.fileno(), msg.decode())).encode()
            for client in clients:
                if conn != client:
                    yield client.send_msg(msg)
        elif cmd == 'joined':
            conn, addr = item
            yield conn.send_msg(('id: %s' % conn.fileno()).encode())
            msg = ('%s joined' % conn.fileno()).encode()
            for client in clients:
                yield client.send_msg(msg)
            clients.add(conn)
        elif cmd == 'left':
            conn, addr = item
            clients.discard(conn)
            msg = ('%s left' % conn.fileno()).encode()
            conn.close()
            for client in clients:
                yield client.send_msg(msg)
        elif cmd == 'terminate':
            break

if __name__ == '__main__':
    asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # host name or IP address of this node is arg1
    if len(sys.argv) > 1:
        host = sys.argv[1]
    else:
        host = ''

    # port to use is arg2
    if len(sys.argv) > 2:
        port = int(sys.argv[2])
    else:
        port = 1234

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(128)

    asyncoro.logger.info('server at %s', str(sock.getsockname()))
    server = asyncoro.Coro(server_proc)
    try:
        while True:
            conn, addr = sock.accept()
            # each connection is handled by a coroutine
            asyncoro.Coro(client_proc, conn, addr, server)
    except:
        pass
    server.send(('terminate'.encode(), None))



# Run 'discoronode.py' program to start processes to execute computations sent
# by this client, along with this program.

# Distributed computing example where this client sends computation to remote
# discoro process to run as remote coroutines. At any time at most one
# computation coroutine is scheduled at a process, as the computation is
# supposed to be CPU heavy (although in this example they are not). Status of
# remote coroutines can be monitored in a web browser at http://127.0.0.1:8181

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *
import asyncoro.httpd


# user computations should be generator functions.
def compute(n, coro=None):
    yield coro.sleep(n)

def client_proc(computation, njobs, coro=None):
    # schedule computation with the scheduler; scheduler accepts one computation
    # at a time, so if scheduler is shared, the computation is queued until it
    # is done with already scheduled computations
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    # run jobs
    for i in range(njobs):
        # computation is supposed to be CPU bound so 'run' is used so at most
        # one computations runs at a server at any time; for mostly idle
        # computations, use 'run_async' to run more than one computation at a
        # server at the same time.
        rcoro = yield computation.run(compute, random.uniform(5, 10))
        if isinstance(rcoro, asyncoro.Coro):
            print('  job %s processed by %s' % (i, rcoro.location))
        else:
            print('rcoro %s failed: %s' % (i, rcoro))

    # wait for all jobs to be done and close computation
    yield computation.close()


if __name__ == '__main__':
    import random, sys, asyncoro.discoro
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is not already running (on a node as a program), start
    # private scheduler:
    Scheduler()
    # send 'compute' generator function; use MinPulseInterval so node status
    # updates are sent more frequently (instead of default 2*MinPulseInterval)
    computation = Computation([compute], timeout=5, pulse_interval=asyncoro.discoro.MinPulseInterval)

    # to illustrate relaying of status messages to multiple coroutines, httpd is
    # also used in this example; this sets computation's status_coro to httpd's status_coro
    httpd = asyncoro.httpd.HTTPServer(computation)
    # run 10 (or given number of) jobs
    asyncoro.Coro(client_proc, computation, 10 if len(sys.argv) < 2 else int(sys.argv[1])).value()
    # shutdown httpd only after computation is closed; alternately, close it in
    # 'client_proc' after the computation is closed.
    httpd.shutdown()



# Run 'discoronode.py' program to start processes to execute computations sent
# by this client, along with this program.

# Distributed computing example where this client sends computation ('compute'
# function)to remote discoro servers to run as remote coroutines and obtain
# results. At any time at most one computation coroutine is scheduled at a
# process, as the computation is supposed to be CPU heavy (although in this
# example they are not).

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *

# this generator function is sent to remote discoro servers to run coroutines
# there
def compute(i, n, coro=None):
    import time
    yield coro.sleep(n)
    raise StopIteration((i, coro.location, time.asctime())) # result of 'compute' is current time

# client (local) coroutine submits computations
def client_proc(computation, njobs, coro=None):
    # schedule computation with the scheduler; scheduler accepts one computation
    # at a time, so if scheduler is shared, the computation is queued until it
    # is done with already scheduled computations
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    # arguments must correspond to arguments for computaiton; multiple arguments
    # (as in this case) can be given as tuples
    args = [(i, random.uniform(2, 5)) for i in range(njobs)]
    results = yield computation.run_results(compute, args)
    # Coroutines may not be executed in the order of given list of args, but
    # results would be in the same order of given list of args
    for result in results:
        print('    result for %d from %s: %s' % result)

    # wait for all jobs to be done and close computation
    yield computation.close()


if __name__ == '__main__':
    import random, sys
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is not already running (on a node as a program), start
    # private scheduler:
    Scheduler()
    # package computation fragments
    computation = Computation([compute])
    # run 10 (or given number of) jobs
    asyncoro.Coro(client_proc, computation, 10 if len(sys.argv) < 2 else int(sys.argv[1]))



#!/usr/bin/python3

"""This file is part of asyncoro; see http://asyncoro.sourceforge.net
for details.

This module provides API for creating distributed communicating
processes. 'Computation' class should be used to package computation
components (Python generator functions, Python functions, files,
classes, modules) and then schedule runs that create remote coroutines
at remote server processes running 'discoronode.py'.

See 'discoro_client*.py' files in 'examples' directory for various use
cases.
"""

import os
import sys
import inspect
import hashlib
import collections
import time
import socket
import shutil
import operator
import functools
import re

import asyncoro.disasyncoro as asyncoro
from asyncoro import Coro, SysCoro, logger
import asyncoro.discoro

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2014-2015 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

__all__ = ['Scheduler', 'Computation', 'DiscoroStatus', 'DiscoroCoroInfo',
           'DiscoroNodeInfo', 'DiscoroNodeAvailInfo', 'DiscoroNodeAllocate']

MsgTimeout = asyncoro.MsgTimeout
MinPulseInterval = 10
MaxPulseInterval = 10 * MinPulseInterval

# status about nodes / servers are sent with this structure
DiscoroStatus = collections.namedtuple('DiscoroStatus', ['status', 'info'])
DiscoroCoroInfo = collections.namedtuple('DiscoroCoroInfo', ['coro', 'args', 'kwargs', 'start_time'])
DiscoroNodeInfo = collections.namedtuple('DiscoroNodeInfo', ['name', 'addr', 'cpus', 'platform',
                                                             'avail_info'])


class DiscoroNodeAvailInfo(object):
    """Node availability status is indicated with this class.  'cpu' is
    available CPU in percent in the range 0 to 100. 0 indicates node is busy
    executing tasks on all CPUs and 100 indicates node is not busy at all.
    """

    def __init__(self, location, cpu, memory, disk, swap):
        self.location = location
        self.cpu = cpu
        self.memory = memory
        self.disk = disk
        self.swap = swap


class DiscoroNodeAllocate(object):
    """Allocation of nodes can be customized by specifying 'node_allocations' of
    Computation with DiscoroNodeAllocate instances.

    'ip_addr' must be a regular expression, e.g., "192\.168\.1\..*" (note that
    "." without escape stands for "any" character), 'cpus' must be minimum
    number of servers running on that node, 'memory' is available memory in
    bytes, 'disk' is available disk space (on the partition where discoronode
    servers are running from), and 'platform' is regular expression to match
    output of"platform.platform()" on that node, e.g., "Linux.*x86_64" to accept
    only nodes that run 64-bit Linux.
    """

    def __init__(self, node, platform='', cpus=0, memory=0, disk=0):
        if node.find('*') < 0:
            try:
                node = socket.gethostbyname(node)
            except:
                node = ''

        if node:
            self.ip_rex = node.replace('.', '\\.').replace('*', '.*')
        else:
            logger.warning('node "%s" is invalid', node)
            self.ip_rex = ''
        self.platform = platform.lower()
        self.cpus = cpus
        self.memory = memory
        self.disk = disk

    def allocate(self, ip_addr, name, platform, cpus, memory, disk):
        """When a node is found, scheduler calls this method with IP address,
        name, CPUs, memory and disk available on that node. This method should
        return a number indicating number of CPUs to use. If return value is 0,
        the node is not used; if the return value is < 0, this allocation is
        ignored (next allocation in the 'node_allocations' list, if any, is
        applied).
        """
        if not re.match(self.ip_rex, ip_addr):
            return -1
        if (self.platform and not re.search(self.platform, platform)):
            return -1
        if ((self.memory and memory and self.memory > memory) or
            (self.disk and disk and self.disk > disk)):
            return 0
        if self.cpus > 0:
            if self.cpus > cpus:
                return 0
            return self.cpus
        elif self.cpus == 0:
            return 0
        else:
            cpus += self.cpus
            if cpus < 0:
                return 0
            return cpus


class Computation(object):
    """Packages components to distribute to remote asyncoro schedulers to create
    (remote) coroutines.
    """

    def __init__(self, components, pulse_interval=(5*MinPulseInterval), node_allocations=[],
                 status_coro=None, node_setup=None, server_setup=None,
                 disable_nodes=False, disable_servers=False, peers_communicate=False):
        """'components' should be a list, each element of which is either a
        module, a (generator or normal) function, path name of a file, a class
        or an object (in which case the code for its class is sent).

        'pulse_interval' is interval (number of seconds) used for heart beat
        messages to check if client / scheduler / server is alive. If the other
        side doesn't reply to 5 heart beat messages, it is treated as dead.
        """

        if pulse_interval < MinPulseInterval or pulse_interval > MaxPulseInterval:
            raise Exception('"pulse_interval" must be at least %s and at most %s' %
                            (MinPulseInterval, MaxPulseInterval))
        if ((not isinstance(node_allocations, list)) or
            any(not isinstance(_, (DiscoroNodeAllocate, str)) for _ in node_allocations)):
            raise Exception('"node_allocations" must be list of DiscoroNodeAllocate instances')
        if status_coro and not isinstance(status_coro, Coro):
            raise Exception('status_coro must be Coro instance')
        if node_setup and not inspect.isgeneratorfunction(node_setup):
            raise Exception('"node_setup" must be a coroutine (generator function)')
        if server_setup and not inspect.isgeneratorfunction(server_setup):
            raise Exception('"server_setup" must be a coroutine (generator function)')
        if (disable_nodes or disable_servers) and not status_coro:
            raise Exception('status_coro must be given when nodes or servers are disabled')

        if not isinstance(components, list):
            components = [components]

        self._code = ''
        self._xfer_funcs = set()
        self._xfer_files = []
        self._auth = None
        self.scheduler = None
        self._pulse_coro = None
        self._pulse_interval = pulse_interval
        self._node_allocations = [node if isinstance(node, DiscoroNodeAllocate)
                                  else DiscoroNodeAllocate(node) for node in node_allocations]
        self.status_coro = status_coro
        if node_setup:
            components.append(node_setup)
            self._node_setup = node_setup.__name__
        else:
            self._node_setup = None
        if server_setup:
            components.append(server_setup)
            self._server_setup = server_setup.__name__
        else:
            self._server_setup = None
        self._peers_communicate = bool(peers_communicate)
        self._disable_nodes = bool(disable_nodes)
        self._disable_servers = bool(disable_servers)

        depends = set()
        cwd = os.getcwd()
        for dep in components:
            if isinstance(dep, str) or inspect.ismodule(dep):
                if inspect.ismodule(dep):
                    name = dep.__file__
                    if name.endswith('.pyc'):
                        name = name[:-1]
                    if not name.endswith('.py'):
                        raise Exception('Invalid module "%s" - must be python source.' % dep)
                    if name.startswith(cwd):
                        dst = os.path.dirname(name[len(cwd):].lstrip(os.sep))
                    elif dep.__package__:
                        dst = dep.__package__.replace('.', os.sep)
                    else:
                        dst = os.path.dirname(dep.__name__.replace('.', os.sep))
                else:
                    name = os.path.abspath(dep)
                    if name.startswith(cwd):
                        dst = os.path.dirname(name[len(cwd):].lstrip(os.sep))
                    else:
                        dst = '.'
                if name in depends:
                    continue
                try:
                    with open(name, 'rb') as fd:
                        pass
                except:
                    raise Exception('File "%s" is not valid' % name)
                self._xfer_files.append((name, dst, os.sep))
                depends.add(name)
            elif (inspect.isgeneratorfunction(dep) or inspect.isfunction(dep) or
                  inspect.isclass(dep) or hasattr(dep, '__class__')):
                if inspect.isgeneratorfunction(dep) or inspect.isfunction(dep):
                    name = dep.__name__
                elif inspect.isclass(dep):
                    name = dep.__name__
                elif hasattr(dep, '__class__') and inspect.isclass(dep.__class__):
                    dep = dep.__class__
                    name = dep.__name__
                if name in depends:
                    continue
                depends.add(name)
                self._xfer_funcs.add(name)
                self._code += '\n' + inspect.getsource(dep).lstrip()
            else:
                raise Exception('Invalid computation: %s' % dep)
        # check code can be compiled
        compile(self._code, '<string>', 'exec')
        # Under Windows discoro server may send objects with '__mp_main__'
        # scope, so make an alias to '__main__'.  Do so even if scheduler is not
        # running on Windows; it is possible the client is not Windows, but a
        # node is.
        if os.name == 'nt' and '__mp_main__' not in sys.modules:
            sys.modules['__mp_main__'] = sys.modules['__main__']

    def schedule(self, location=None, timeout=None):
        """Schedule computation for execution. Must be used with 'yield' as
        'result = yield compute.schedule()'. If scheduler is executing other
        computations, this will block until scheduler processes them
        (computations are processed in the order submitted).
        """

        if self._auth is not None:
            raise StopIteration(0)
        self._auth = ''
        if self.status_coro is not None and not isinstance(self.status_coro, Coro):
            raise StopIteration(-1)

        self.scheduler = yield SysCoro.locate('discoro_scheduler', location=location,
                                              timeout=MsgTimeout)
        if not isinstance(self.scheduler, Coro):
            raise StopIteration(-1)

        def _schedule(self, coro=None):
            self._pulse_coro = SysCoro(self._pulse_proc)
            msg = {'req': 'schedule', 'computation': asyncoro.serialize(self), 'client': coro}
            self.scheduler.send(msg)
            self._auth = yield coro.receive(timeout=MsgTimeout)
            if not isinstance(self._auth, str):
                logger.debug('Could not send computation to scheduler %s: %s',
                             self.scheduler, self._auth)
                raise StopIteration(-1)
            SysCoro.scheduler().atexit(10, lambda: SysCoro(self.close))
            if coro.location != self.scheduler.location:
                for xf, dst, sep in self._xfer_files:
                    drive, xf = os.path.splitdrive(xf)
                    if xf.startswith(sep):
                        xf = os.path.join(os.sep, *(xf.split(sep)))
                    else:
                        xf = os.path.join(*(xf.split(sep)))
                    xf = drive + xf
                    dst = os.path.join(self._auth, os.path.join(*(dst.split(sep))))
                    if (yield asyncoro.AsynCoro.instance().send_file(
                       self.scheduler.location, xf, dir=dst, timeout=MsgTimeout)) < 0:
                        logger.warning('Could not send file "%s" to scheduler', xf)
                        yield self.close()
                        raise StopIteration(-1)
            msg = {'req': 'await', 'auth': self._auth, 'client': coro}
            self.scheduler.send(msg)
            resp = yield coro.receive()
            if (isinstance(resp, dict) and resp.get('auth') == self._auth and
               resp.get('resp') == 'scheduled'):
                raise StopIteration(0)
            else:
                yield self.close()
                raise StopIteration(-1)

        yield Coro(_schedule, self).finish()

    def nodes(self):
        """Get list of addresses of nodes initialized for this computation. Must
        be used with 'yield' as 'yield compute.nodes()'.
        """

        def _nodes(self, coro=None):
            msg = {'req': 'nodes', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=MsgTimeout)) == 1:
                yield coro.receive(MsgTimeout)
            else:
                raise StopIteration([])

        yield Coro(_nodes, self).finish()

    def servers(self):
        """Get list of Location instances of servers initialized for this
        computation. Must be used with 'yield' as 'yield compute.servers()'.
        """

        def _servers(self, coro=None):
            msg = {'req': 'servers', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=MsgTimeout)) == 1:
                yield coro.receive(MsgTimeout)
            else:
                raise StopIteration([])

        yield Coro(_servers, self).finish()

    def close(self, await_async=False):
        """Close computation. Must be used with 'yield' as 'yield
        compute.close()'.
        """

        def _close(self, done, coro=None):
            msg = {'req': 'close_computation', 'auth': self._auth, 'client': coro,
                   'await_async': bool(await_async)}
            yield self.scheduler.deliver(msg, timeout=MsgTimeout)
            msg = yield coro.receive()
            if msg != 'closed':
                logger.warning('%s: closing computation failed?', self._auth)
            self._auth = None
            if self._pulse_coro:
                yield self._pulse_coro.send('quit')
                self._pulse_coro = None
            done.set()

        if self._auth:
            done = asyncoro.Event()
            SysCoro(_close, self, done)
            yield done.wait()

    def run_at(self, where, gen, *args, **kwargs):
        """Must be used with 'yield' as

        'rcoro = yield computation.run_at(where, gen, ...)'

        Run given generator function 'gen' with arguments 'args' and 'kwargs' at
        remote server 'where'.  If the request is successful, 'rcoro' will be a
        (remote) coroutine; check result with 'isinstance(rcoro,
        asyncoro.Coro)'. The generator is expected to be (mostly) CPU bound and
        until this is finished, another CPU bound coroutine will not be
        submitted at same server.

        If 'where' is a string, it is assumed to be IP address of a node, in
        which case the coroutine is scheduled at that node on a server at that
        node. If 'where' is a Location instance, it is assumed to be server
        location in which case the coroutine is scheduled at that server.

        'gen' must be generator function, as it is used to run coroutine at
        remote location.
        """
        yield self._run_request('run_async', where, 1, gen, *args, **kwargs)

    def run(self, gen, *args, **kwargs):
        """Run CPU bound coroutine at any remote server; see 'run_at'
        above.
        """
        yield self._run_request('run_async', None, 1, gen, *args, **kwargs)

    def run_result_at(self, where, gen, *args, **kwargs):
        """Must be used with 'yield' as

        'rcoro = yield computation.run_result_at(where, gen, ...)'

        Whereas 'run_at' and 'run' return remote coroutine instance,
        'run_result_at' and 'run_result' wait until remote coroutine is
        finished and return the result of that remote coroutine (i.e., either
        the value of 'StopIteration' or the last value 'yield'ed).

        'where', 'gen', 'args', 'kwargs' are as explained in 'run_at'.
        """
        yield self._run_request('run_result', where, 1, gen, *args, **kwargs)

    def run_result(self, gen, *args, **kwargs):
        """Run CPU bound coroutine at any remote server and return result of
        that coroutine; see 'run_result_at' above.
        """
        yield self._run_request('run_result', None, 1, gen, *args, **kwargs)

    def run_async_at(self, where, gen, *args, **kwargs):
        """Must be used with 'yield' as

        'rcoro = yield computation.run_async_at(where, gen, ...)'

        Run given generator function 'gen' with arguments 'args' and 'kwargs' at
        remote server 'where'.  If the request is successful, 'rcoro' will be a
        (remote) coroutine; check result with 'isinstance(rcoro,
        asyncoro.Coro)'. The generator is supposed to be (mostly) I/O bound and
        not consume CPU time. Unlike other 'run' variants, coroutines created
        with 'async' are not "tracked" by scheduler (see online documentation for
        more details).

        If 'where' is a string, it is assumed to be IP address of a node, in
        which case the coroutine is scheduled at that node on a server at that
        node. If 'where' is a Location instance, it is assumed to be server
        location in which case the coroutine is scheduled at that server.

        'gen' must be generator function, as it is used to run coroutine at
        remote location.
        """
        yield self._run_request('run_async', where, 0, gen, *args, **kwargs)

    def run_async(self, gen, *args, **kwargs):
        """Run I/O bound coroutine at any server; see 'run_async_at'
        above.
        """
        yield self._run_request('run_async', None, 0, gen, *args, **kwargs)

    def run_results(self, gen, iter):
        """Must be used with 'yield', as for example,
        'results = yield scheduler.map_results(generator, list_of_tuples)'.

        Execute generator 'gen' with arguments from given iterable. The return
        value is list of results that correspond to executing 'gen' with
        arguments in iterable in the same order.
        """
        coros = []
        append_coro = coros.append
        for params in iter:
            if not isinstance(params, tuple):
                if hasattr(params, '__iter__'):
                    params = tuple(params)
                else:
                    params = (params,)
            append_coro(Coro(self.run_result, gen, *params))
        results = [None] * len(coros)
        for i, coro in enumerate(coros):
            results[i] = yield coro.finish()
        raise StopIteration(results)

    def enable_node(self, ip_addr, *setup_args):
        """If computation disabled nodes (with 'disabled_nodes=True' when
        Computation is constructed), nodes are not automatically used by the
        scheduler until nodes are enabled with 'enable_node'.

        'ip_addr' must be either IP address or host name of the node to be
        enabled.

        'setup_args' is arguments passed to 'node_setup' function specific to
        that node. If 'node_setup' succeeds (i.e., finishes with value 0), the
        node is used for computations.
        """
        if self.scheduler:
            if isinstance(ip_addr, asyncoro.Location):
                ip_addr = ip_addr.addr
            self.scheduler.send({'req': 'enable_node', 'auth': self._auth, 'addr': ip_addr,
                                 'setup_args': setup_args})

    def enable_server(self, location, *setup_args):
        """If computation disabled servers (with 'disabled_servers=True' when
        Computation is constructed), servers are not automatically used by the
        scheduler until they are enabled with 'enable_server'.

        'location' must be Location instance of the server to be enabled.

        'setup_args' is arguments passed to 'server_setup' function specific to
        that server. If 'server_setup' succeeds (i.e., finishes with value 0), the
        server is used for computations.
        """
        if self.scheduler:
            self.scheduler.send({'req': 'enable_server', 'auth': self._auth, 'server': location,
                                 'setup_args': setup_args})

    def _run_request(self, request, where, cpu, gen, *args, **kwargs):
        """Internal use only.
        """
        if isinstance(gen, str):
            name = gen
        else:
            name = gen.__name__

        if name in self._xfer_funcs:
            code = None
        else:
            # if not inspect.isgeneratorfunction(gen):
            #     logger.warning('"%s" is not a valid generator function', name)
            #     raise StopIteration([])
            code = inspect.getsource(gen).lstrip()

        def _run_req(coro=None):
            msg = {'req': 'job', 'auth': self._auth,
                   'job': _DiscoroJob_(request, coro, name, where, cpu, code, args, kwargs)}
            if (yield self.scheduler.deliver(msg, timeout=MsgTimeout)) == 1:
                reply = yield coro.receive()
                if isinstance(reply, Coro):
                    if self.status_coro:
                        msg = DiscoroCoroInfo(reply, args, kwargs, time.time())
                        self.status_coro.send(DiscoroStatus(Scheduler.CoroCreated, msg))
                if not request.endswith('async'):
                    reply = yield coro.receive()
            else:
                reply = None
            raise StopIteration(reply)

        yield Coro(_run_req).finish()

    def _pulse_proc(self, coro=None):
        """For internal use only.
        """
        coro.set_daemon()
        last_pulse = time.time()
        while 1:
            msg = yield coro.receive(timeout=(2 * self._pulse_interval))
            if msg == 'pulse':
                last_pulse = time.time()
            elif msg == 'quit':
                break
            elif msg is None and (time.time() - last_pulse) > (10 * self._pulse_interval):
                logger.warning('scheduler may have gone away!')
            else:
                logger.debug('ignoring invalid pulse message')
        self._pulse_coro = None

    def __getstate__(self):
        state = {}
        for attr in ['_code', '_xfer_funcs', '_xfer_files', '_auth',  'scheduler', 'status_coro',
                     '_pulse_interval', '_pulse_coro', '_node_allocations',
                     '_node_setup', '_server_setup', '_disable_nodes', '_disable_servers',
                      '_peers_communicate']:
            state[attr] = getattr(self, attr)
        return state

    def __setstate__(self, state):
        for attr, value in state.items():
            setattr(self, attr, value)


class _DiscoroJob_(object):
    """Internal use only.
    """
    __slots__ = ('request', 'client', 'name', 'where', 'cpu', 'code', 'args', 'kwargs', 'done')

    def __init__(self, request, client, name, where, cpu, code, args=None, kwargs=None):
        self.request = request
        self.client = client
        self.name = name
        self.where = where
        self.cpu = cpu
        self.code = code
        self.args = asyncoro.serialize(args)
        self.kwargs = asyncoro.serialize(kwargs)
        self.done = None


class Scheduler(object, metaclass=asyncoro.Singleton):

    # status indications ('status' attribute of DiscoroStatus)
    NodeDiscovered = 1
    NodeInitialized = 2
    NodeClosed = 3
    NodeIgnore = 4
    NodeDisconnected = 5

    ServerDiscovered = 11
    ServerInitialized = 12
    ServerClosed = 13
    ServerIgnore = 14
    ServerDisconnected = 15

    CoroCreated = 20
    ComputationScheduled = 23
    ComputationClosed = 25

    _instance = None

    """This class is for use by Computation class (see below) only.  Other than
    the status indications above, none of its attributes are to be accessed
    directly.
    """

    __status_coro = None

    class _Node(object):

        def __init__(self, name, addr):
            self.name = name
            self.addr = addr
            self.cpus_used = 0
            self.platform = None
            self.avail_info = None
            self.servers = {}
            self.disabled_servers = {}
            self.load = 0.0
            self.status = Scheduler.NodeClosed
            self.coro = None
            self.last_pulse = time.time()
            self.lock = asyncoro.Lock()
            self.avail = asyncoro.Event()
            self.avail.clear()

    class _Server(object):

        def __init__(self, name, location):
            self.name = name
            self.coro = None
            self.status = Scheduler.ServerClosed
            self.rcoros = {}
            self.xfer_files = []
            self.askew_results = {}
            self.avail = asyncoro.Event()
            self.avail.clear()
            self.scheduler = Scheduler._instance

        def run(self, job, computation, node):
            def _run(self, coro=None):
                self.coro.send({'req': 'run', 'auth': computation._auth, 'job': job, 'client': coro})
                rcoro = yield coro.receive(timeout=MsgTimeout)
                # currently fault-tolerancy is not supported, so clear job's
                # args to save space
                job.args = job.kwargs = None
                if isinstance(rcoro, Coro):
                    # TODO: keep func too for fault-tolerance
                    job.done = asyncoro.Event()
                    self.rcoros[rcoro] = (rcoro, job)
                    if self.askew_results:
                        msg = self.askew_results.pop(rcoro, None)
                        if msg:
                            Scheduler.__status_coro.send(msg)
                else:
                    logger.debug('failed to create rcoro: %s', rcoro)
                    if job.cpu:
                        self.avail.set()
                        node.cpus_used -= 1
                        node.load = float(node.cpus_used) / len(node.servers)
                        self.scheduler._avail_nodes.add(node)
                        self.scheduler._nodes_avail.set()
                        node.avail.set()
                raise StopIteration(rcoro)

            rcoro = yield SysCoro(_run, self).finish()
            job.client.send(rcoro)

    def __init__(self, **kwargs):
        self.__class__._instance = self
        self._nodes = {}
        self._disabled_nodes = {}
        self._avail_nodes = set()
        self._nodes_avail = asyncoro.Event()
        self._nodes_avail.clear()

        self._cur_computation = None
        self.__cur_client_auth = None
        self.__cur_node_allocations = []
        self.__pulse_interval = kwargs.pop('pulse_interval', MaxPulseInterval)
        self.__ping_interval = kwargs.pop('ping_interval', 0)
        self.__zombie_period = kwargs.pop('zombie_period', 100 * MaxPulseInterval)
        self._node_port = kwargs.pop('discoronode_port', 51351)
        self.__server_locations = set()
        self.__job_scheduler_coro = None

        kwargs['name'] = 'discoro_scheduler'
        clean = kwargs.pop('clean', False)
        nodes = kwargs.pop('nodes', [])
        self.asyncoro = asyncoro.AsynCoro.instance(**kwargs)
        self.__dest_path = os.path.join(self.asyncoro.dest_path, 'discoro', 'discoroscheduler')
        if clean:
            shutil.rmtree(self.__dest_path)
        self.asyncoro.dest_path = self.__dest_path

        self.__computation_sched_event = asyncoro.Event()
        self.__computation_scheduler_coro = SysCoro(self.__computation_scheduler_proc, nodes)
        self.__client_coro = SysCoro(self.__client_proc)
        self.__timer_coro = SysCoro(self.__timer_proc)
        Scheduler.__status_coro = self.__status_coro = SysCoro(self.__status_proc)
        self.__client_coro.register('discoro_scheduler')
        self.asyncoro.discover_peers(port=self._node_port)

    def status(self):
        pending = sum(node.cpus_used for node in self._nodes.values())
        servers = functools.reduce(operator.add, [list(node.servers.keys())
                                                  for node in self._nodes.values()], [])
        return {'Client': self._cur_computation._pulse_coro.location if self._cur_computation else '',
                'Pending': pending, 'Nodes': list(self._nodes.keys()), 'Servers': servers
                }

    def print_status(self):
        status = self.status()
        print('')
        print('  Client: %s' % status['Client'])
        print('  Pending: %s' % status['Pending'])
        print('  nodes: %s' % len(status['Nodes']))
        print('  servers: %s' % len(status['Servers']))

    def __status_proc(self, coro=None):
        coro.set_daemon()
        coro.register('discoro_status')
        self.asyncoro.peer_status(coro)
        while 1:
            msg = yield coro.receive()
            now = time.time()
            if isinstance(msg, asyncoro.MonitorException):
                rcoro = msg.args[0]
                if not isinstance(rcoro, Coro):
                    logger.warning('ignoring invalid rcoro %s', type(rcoro))
                    continue
                node = self._nodes.get(rcoro.location.addr, None)
                if not node:
                    logger.warning('node %s is invalid', rcoro.location.addr)
                    continue
                server = node.servers.get(rcoro.location, None)
                if not server:
                    logger.warning('server "%s" is invalid', rcoro.location)
                    continue
                node.last_pulse = now
                info = server.rcoros.pop(rcoro, None)
                if not info:
                    # Due to 'yield' used to create rcoro, scheduler may not
                    # have updated self._rcoros before the coroutine's
                    # MonitorException is received, so put it in
                    # 'askew_results'. The scheduling coroutine will resend it
                    # when it receives rcoro
                    server.askew_results[rcoro] = msg
                    continue
                # assert isinstance(info[1], _DiscoroJob_)
                job = info[1]
                if job.cpu:
                    node.cpus_used -= 1
                    node.load = float(node.cpus_used) / len(node.servers)
                    self._avail_nodes.add(node)
                    self._nodes_avail.set()
                    node.avail.set()
                    server.avail.set()
                if job.request.endswith('async'):
                    if job.done:
                        job.done.set()
                else:
                    job.client.send(msg.args[1][1])
                if self._cur_computation and self._cur_computation.status_coro:
                    if len(msg.args) > 2:
                        msg.args = (msg.args[0], msg.args[1])
                    self._cur_computation.status_coro.send(msg)

            elif isinstance(msg, asyncoro.PeerStatus):
                if msg.status == asyncoro.PeerStatus.Online:
                    if msg.name.endswith('_proc-0'):
                        SysCoro(self.__discover_node, msg)
                else:
                    # msg.status == asyncoro.PeerStatus.Offline
                    node = server = None
                    node = self._nodes.get(msg.location.addr, None)
                    if node:
                        server = node.servers.pop(msg.location, None)
                        if server:
                            SysCoro(self.__close_server, server, self._cur_computation)
                        elif node.coro and node.coro.location == msg.location:
                            # TODO: inform scheduler / client
                            self._nodes.pop(msg.location.addr)
                        else:
                            node = None
                        # if not node.servers:
                        #     self._nodes.pop(msg.location.addr)
                    if ((not server and not node) and self._cur_computation and
                        self._cur_computation._pulse_coro.location == msg.location):
                        logger.warning('Client %s terminated; closing computation %s',
                                       msg.location, self.__cur_client_auth)
                        SysCoro(self.__close_computation)

            elif isinstance(msg, dict):  # message from a node's server
                status = msg.get('status', None)
                if status == 'pulse':
                    location = msg.get('location', None)
                    if not isinstance(location, asyncoro.Location):
                        continue
                    node = self._nodes.get(location.addr, None)
                    if node:
                        node.last_pulse = now
                        node_status = msg.get('node_status', None)
                        if (node_status and self._cur_computation and
                           self._cur_computation.status_coro):
                            self._cur_computation.status_coro.send(node_status)

                elif status == Scheduler.ServerDiscovered:
                    rcoro = msg.get('coro', None)
                    if not isinstance(rcoro, asyncoro.Coro):
                        continue
                    node = self._nodes.get(rcoro.location.addr, None)
                    if not node:
                        node = self._disabled_nodes.get(rcoro.location.addr, None)
                    if not node or (node.status != Scheduler.NodeInitialized and
                                    node.status != Scheduler.NodeDiscovered):
                        continue
                    server = node.servers.get(rcoro.location, None)
                    if not server:
                        server = node.disabled_servers.get(rcoro.location, None)
                    if server:
                        continue
                    server = Scheduler._Server(msg.get('name', None), rcoro.location)
                    server.coro = rcoro
                    server.status = Scheduler.ServerDiscovered
                    node.disabled_servers[rcoro.location] = server
                    if self._cur_computation and self._cur_computation.status_coro:
                        info = DiscoroStatus(server.status, server.coro.location)
                        self._cur_computation.status_coro.send(info)

                elif status == Scheduler.ServerInitialized:
                    rcoro = msg.get('coro', None)
                    if not isinstance(rcoro, asyncoro.Coro):
                        continue
                    if (not self._cur_computation or
                        self._cur_computation._auth != msg.get('auth', None)):
                        continue
                    node = self._nodes.get(rcoro.location.addr, None)
                    if not node:
                        node = self._disabled_nodes.get(rcoro.location.addr, None)
                    if not node or (node.status != Scheduler.NodeInitialized and
                                    node.status != Scheduler.NodeDiscovered):
                        continue
                    server = node.disabled_servers.pop(rcoro.location, None)
                    if not server:
                        server = node.servers.get(rcoro.location, None)
                    if server:
                        if server.status != Scheduler.ServerDiscovered:
                            continue
                    else:
                        server = Scheduler._Server(msg.get('name', None), rcoro.location)
                        server.coro = rcoro

                    server.status = Scheduler.ServerInitialized
                    node.servers[rcoro.location] = server
                    node.last_pulse = now
                    if node.status == Scheduler.NodeInitialized:
                        if self._cur_computation.status_coro:
                            self._cur_computation.status_coro.send(
                                DiscoroStatus(server.status, server.coro.location))
                        server.avail.set()
                        node.avail.set()
                        self._avail_nodes.add(node)
                        self._nodes_avail.set()
                    if self._cur_computation._peers_communicate:
                        server.coro.send({'req': 'peers', 'auth': self._cur_computation._auth,
                                          'peers': list(self.__server_locations)})
                        self.__server_locations.add(server.coro.location)

                elif status in (Scheduler.ServerClosed, Scheduler.ServerDisconnected):
                    location = msg.get('location', None)
                    if not isinstance(location, asyncoro.Location):
                        continue
                    node = self._nodes.get(location.addr, None)
                    if not node:
                        continue
                    if status == Scheduler.ServerClosed:
                        server = node.servers.get(location, None)
                        if server:
                            server.status = Scheduler.ServerIgnore
                    else:
                        server = node.servers.pop(location, None)
                        if not node.servers:
                            node.status = Scheduler.NodeIgnore
                    if not server:
                        continue
                    if not node.servers:
                        self._avail_nodes.discard(node)
                        if not self._avail_nodes:
                            self._nodes_avail.clear()
                        # self._nodes.pop(node.addr, None)
                    if self._cur_computation:
                        if self._cur_computation._peers_communicate:
                            self.__server_locations.discard(server.coro.location)
                            # TODO: inform other servers
                        if self._cur_computation.status_coro:
                            self._cur_computation.status_coro.send(DiscoroStatus(status, location))
                            # if not node.servers:
                            #     info = DiscoroNodeInfo(node.name, node.addr, node.cpus,
                            #                            node.platform, node.avail_info)
                            #     self._cur_computation.status_coro.send(
                            #         Scheduler.NodeClosed, DiscoroStatus(node.status, info))

                else:
                    logger.warning('Ignoring invalid status message: %s', status)
            else:
                logger.warning('invalid status message ignored')

    def __node_allocate(self, node):
        for node_allocate in self.__cur_node_allocations:
            cpus = node_allocate.allocate(node.addr, node.name, node.platform, node.cpus,
                                          node.avail_info.memory, node.avail_info.disk)
            if cpus < 0:
                continue
            return min(cpus, node.cpus)
        return node.cpus

    def __get_node_info(self, node, coro=None):
        if not node.coro:
            logger.warning('Node %s is not valid', node.addr)
            raise StopIteration
        # assert node.addr in self._disabled_nodes
        node.coro.send({'req': 'discoro_node_info', 'client': coro})
        node_info = yield coro.receive(timeout=MsgTimeout)
        if not node_info:
            asyncoro.Coro(asyncoro.AsynCoro.instance().close_peer, node.coro.location)
            raise StopIteration
        node.name = node_info.name
        node.cpus = node_info.cpus
        node.platform = node_info.platform.lower()
        node.avail_info = node_info.avail_info
        if self._cur_computation:
            yield self.__init_node(node, coro=coro)

    def __init_node(self, node, setup_args=(), coro=None):
        computation = self._cur_computation
        if not computation or not node.coro:
            raise StopIteration(-1)
        # this coroutine may be invoked in two different paths (when a node is
        # found right after computation is already scheduled, and when
        # computation is scheduled right after a node is found). To prevent
        # concurrent execution (that may reserve / initialize same node more
        # than once), lock is used
        yield node.lock.acquire()
        if node.status == Scheduler.NodeInitialized:
            node.lock.release()
            raise StopIteration(0)

        if node.status == Scheduler.NodeClosed:
            cpus = self.__node_allocate(node)
            if not cpus:
                node.status = Scheduler.NodeIgnore
                node.lock.release()
                raise StopIteration(0)

            node.coro.send({'req': 'reserve', 'cpus': cpus, 'auth': computation._auth,
                            'status_coro': self.__status_coro, 'client': coro,
                            'computation_location': computation._pulse_coro.location})
            cpus = yield coro.receive(timeout=MsgTimeout)
            if not cpus:
                logger.warning('setup of %s failed', node.coro)
                node_coro, node.coro = node.coro, None
                node.lock.release()
                self._disabled_nodes.pop(node.addr, None)
                yield asyncoro.AsynCoro.instance().close_peer(node_coro.location)
                raise StopIteration(-1)

            if computation != self._cur_computation:
                node.status = Scheduler.NodeClosed
                node.coro.send({'req': 'release', 'auth': computation._auth, 'client': None})
                node.lock.release()
                raise StopIteration(-1)

            node.status = Scheduler.NodeDiscovered
            if self._cur_computation and self._cur_computation.status_coro:
                info = DiscoroNodeInfo(node.name, node.addr, node.cpus, node.platform,
                                       node.avail_info)
                self._cur_computation.status_coro.send(DiscoroStatus(node.status, info))
            if self._cur_computation._disable_nodes:
                node.lock.release()
                raise StopIteration(0)

        for xf, dst, sep in computation._xfer_files:
            reply = yield self.asyncoro.send_file(node.coro.location, xf, dir=dst,
                                                  timeout=MsgTimeout)
            if reply < 0 or computation != self._cur_computation:
                logger.debug('failed to transfer file %s: %s', xf, reply)
                node.status = Scheduler.NodeClosed
                node.coro.send({'req': 'release', 'auth': computation._auth, 'client': None})
                node.lock.release()
                raise StopIteration(-1)

        node.coro.send({'req': 'computation', 'computation': computation, 'auth': computation._auth,
                        'setup_args': setup_args, 'client': coro})
        cpus = yield coro.receive(timeout=MsgTimeout)
        if not cpus or computation != self._cur_computation:
            node.status = Scheduler.NodeClosed
            node.coro.send({'req': 'release', 'auth': computation._auth, 'client': None})
            node.lock.release()
            raise StopIteration(-1)
        node.cpus = cpus
        node.status = Scheduler.NodeInitialized
        self._disabled_nodes.pop(node.addr, None)
        self._nodes[node.addr] = node
        if computation.status_coro:
            info = DiscoroNodeInfo(node.name, node.addr, node.cpus, node.platform, node.avail_info)
            computation.status_coro.send(DiscoroStatus(node.status, info))
        node.lock.release()
        if node.servers:
            for server in node.servers.values():
                server.avail.set()
                if computation.status_coro:
                    computation.status_coro.send(DiscoroStatus(server.status, server.coro.location))
            node.avail.set()
            self._avail_nodes.add(node)
            self._nodes_avail.set()

    def __discover_node(self, msg, coro=None):
        for _ in range(10):
            rcoro = yield Coro.locate('discoro_node', location=msg.location, timeout=MsgTimeout)
            if not isinstance(rcoro, Coro):
                yield coro.sleep(0.1)
                continue
            node = self._nodes.get(msg.location.addr, None)
            if not node:
                node = self._disabled_nodes.get(msg.location.addr, None)
            if node and node.coro == rcoro:
                raise StopIteration

            if not node:
                node = Scheduler._Node(msg.name, msg.location.addr)
                self._disabled_nodes[msg.location.addr] = node
            node.coro = rcoro
            yield self.__get_node_info(node, coro=coro)
            raise StopIteration

    def __timer_proc(self, coro=None):
        coro.set_daemon()
        node_check = client_pulse = last_ping = time.time()
        async_scheduler = coro.scheduler()
        while 1:
            try:
                msg = yield coro.receive(timeout=self.__pulse_interval)
            except GeneratorExit:
                break
            now = time.time()
            if self.__cur_client_auth:
                if (now - client_pulse) > self.__pulse_interval:
                    if self._cur_computation._pulse_coro.send('pulse') == 0:
                        client_pulse = now
                    elif ((now - client_pulse) > self.__zombie_period):
                        logger.warning('Closing zombie computation %s', self.__cur_client_auth)
                        SysCoro(self.__close_computation)

                if (now - node_check) > self.__zombie_period:
                    node_check = now
                    for node in self._nodes.values():
                        if (node.status != Scheduler.NodeInitialized and
                            node.status != Scheduler.NodeDiscovered):
                            continue
                        if (now - node.last_pulse) > self.__zombie_period:
                            logger.warning('discoro node %s is zombie!', node.addr)
                            SysCoro(self.__close_node, node, self._cur_computation)

                    if not self._cur_computation._disable_nodes:
                        for node in self._disabled_nodes.values():
                            if node.coro:
                                SysCoro(self.__init_node, node)

            if self.__ping_interval and ((now - last_ping) > self.__ping_interval):
                last_ping = now
                async_scheduler.discover_peers(port=self._node_port)

    @staticmethod
    def auth_code():
        # TODO: use uuid?
        return hashlib.sha1(''.join(hex(_)[2:] for _ in os.urandom(10)).encode()).hexdigest()

    def __computation_scheduler_proc(self, nodes, coro=None):
        coro.set_daemon()
        for node in nodes:
            yield asyncoro.AsynCoro.instance().peer(node, broadcast=True)
        while 1:
            if self._cur_computation:
                self.__computation_sched_event.clear()
                yield self.__computation_sched_event.wait()
                continue

            self._cur_computation, client = yield coro.receive()
            if self.__job_scheduler_coro:
                self.__job_scheduler_coro.terminate()
            self.__pulse_interval = self._cur_computation._pulse_interval

            self.__cur_client_auth = self._cur_computation._auth
            self._cur_computation._auth = Scheduler.auth_code()
            self.__cur_node_allocations = self._cur_computation._node_allocations
            self._cur_computation._node_allocations = []

            self._disabled_nodes.update(self._nodes)
            self._avail_nodes.clear()
            self._nodes_avail.clear()
            for node in self._disabled_nodes.values():
                node.disabled_servers.update(node.servers)
                node.servers.clear()
                node.avail.clear()
                for server in node.disabled_servers.values():
                    server.avail.clear()
            logger.debug('Computation %s / %s scheduled', self.__cur_client_auth,
                         self._cur_computation._auth)
            self.__job_scheduler_coro = SysCoro(self.__job_scheduler_proc)
            msg = {'resp': 'scheduled', 'auth': self.__cur_client_auth}
            if (yield client.deliver(msg, timeout=MsgTimeout)) != 1:
                logger.warning('client not reachable?')
                self._cur_client_auth = None
                self._cur_computation = None
                continue
            for node in self._disabled_nodes.values():
                SysCoro(self.__get_node_info, node)
            self.__timer_coro.send(None)
        self.__computation_scheduler_coro = None

    def __job_scheduler_proc(self, coro=None):
        coro.set_daemon()
        while 1:
            job = yield coro.receive()
            if (not isinstance(job, asyncoro.discoro._DiscoroJob_) or
                not isinstance(job.client, Coro)):
                logger.warning('Ignoring invalid client job request')
                continue
            cpu = job.cpu
            where = job.where
            if not where:
                if cpu:
                    while not self._avail_nodes:
                        # self._nodes_avail.clear()
                        yield self._nodes_avail.wait()
                    node = None
                    load = None
                    for host in self._avail_nodes:
                        if host.avail.is_set() and (load is None or host.load < load):
                            node = host
                            load = host.load
                else:
                    while not self._nodes:
                        # self._nodes_avail.clear()
                        yield self._nodes_avail.wait()
                    node = None
                    load = None
                    for host in self._nodes.values():
                        if host.avail.is_set() and (load is None or host.load < load):
                            node = host
                            load = host.load
                server = None
                load = None
                for proc in node.servers.values():
                    if cpu:
                        if proc.avail.is_set() and (load is None or len(proc.rcoros) < load):
                            server = proc
                            load = len(proc.rcoros)
                    elif (load is None or len(proc.rcoros) < load):
                        server = proc
                        load = len(proc.rcoros)
                if not server:
                    job.client.send(None)
                    raise StopIteration
                if cpu:
                    server.avail.clear()
                    node.cpus_used += 1
                    node.load = float(node.cpus_used) / len(node.servers)
                    if node.cpus_used == len(node.servers):
                        node.avail.clear()
                        self._avail_nodes.discard(node)
                        if not self._avail_nodes:
                            self._nodes_avail.clear()
                yield server.run(job, self._cur_computation, node)
            elif isinstance(where, str):
                node = self._nodes.get(where, None)
                if not node:
                    job.client.send(None)
                    raise StopIteration
                server = None
                load = None
                for proc in node.servers.values():
                    if cpu:
                        if proc.avail.is_set() and (load is None or len(proc.rcoros) < load):
                            server = proc
                            load = len(proc.rcoros)
                    elif (load is None or len(proc.rcoros) < load):
                        server = proc
                        load = len(proc.rcoros)
                if not server:
                    job.client.send(None)
                    raise StopIteration
                if cpu:
                    server.avail.clear()
                    node.cpus_used += 1
                    node.load = float(node.cpus_used) / len(node.servers)
                    if node.cpus_used == len(node.servers):
                        node.avail.clear()
                        self._avail_nodes.discard(node)
                        if not self._avail_nodes:
                            self._nodes_avail.clear()
                yield server.run(job, self._cur_computation, node)
            elif isinstance(where, asyncoro.Location):
                node = self._nodes.get(where.addr)
                if not node:
                    job.client.send(None)
                    raise StopIteration
                server = node.servers.get(where)
                if not server:
                    job.client.send(None)
                    raise StopIteration
                if cpu:
                    if server.rcoros:
                        yield server.avail.wait()
                    server.avail.clear()
                    node.cpus_used += 1
                    node.load = float(node.cpus_used) / len(node.servers)
                    if node.cpus_used == len(node.servers):
                        node.avail.clear()
                        self._avail_nodes.discard(node)
                        if not self._avail_nodes:
                            self._nodes_avail.clear()
                yield server.run(job, self._cur_computation, node)
            else:
                job.client.send(None)

    def __client_proc(self, coro=None):
        coro.set_daemon()
        computations = {}

        while 1:
            msg = yield coro.receive()
            if not isinstance(msg, dict):
                continue
            req = msg.get('req', None)
            auth = msg.get('auth', None)
            if self.__cur_client_auth != auth:
                if req == 'schedule' or req == 'await':
                    pass
                else:
                    continue

            if req == 'job':
                self.__job_scheduler_coro.send(msg['job'])
                continue

            client = msg.get('client', None)

            if req == 'enable_server':
                loc = msg.get('server', None)
                if not loc:
                    continue
                node = self._nodes.get(loc.addr, None)
                if not node:
                    continue
                server = node.disabled_servers.get(loc, None)
                if not server or not server.coro or server.status != Scheduler.ServerDiscovered:
                    continue
                args = msg.get('setup_args', ())
                server.coro.send({'req': 'enable_server', 'setup_args': args,
                                  'auth': self._cur_computation._auth})

            elif req == 'enable_node':
                addr = msg.get('addr', None)
                if not addr:
                    continue
                node = self._disabled_nodes.pop(addr, None)
                if not node:
                    continue
                if not node or not node.coro or node.status != Scheduler.NodeDiscovered:
                    continue
                setup_args = msg.get('setup_args', ())
                self._nodes[node.addr] = node
                SysCoro(self.__init_node, node, setup_args=setup_args)

            elif req == 'nodes':
                if isinstance(client, Coro):
                    nodes = [node.addr for node in self._nodes.values()
                             if node.status == Scheduler.NodeInitialized]
                    client.send(nodes)

            elif req == 'servers':
                if isinstance(client, Coro):
                    servers = [server.coro.location for node in self._nodes.values()
                               if node.status == Scheduler.NodeInitialized
                               for server in node.servers.values()
                               # if server.status == Scheduler.ServerInitialized
                               ]
                    client.send(servers)

            elif req == 'schedule':
                if not isinstance(client, Coro):
                    logger.warning('Ignoring invalid client request "%s"', req)
                    continue
                try:
                    computation = asyncoro.deserialize(msg['computation'])
                    assert isinstance(computation, Computation) or \
                        computation.__class__.__name__ == 'Computation'
                    assert isinstance(computation._pulse_coro, Coro)
                    if computation._pulse_coro.location == self.asyncoro.location:
                        computation._pulse_coro._id = int(computation._pulse_coro._id)
                        if computation.status_coro:
                            computation.status_coro._id = int(computation.status_coro._id)
                    assert isinstance(computation._pulse_interval, (float, int))
                    assert (MinPulseInterval <= computation._pulse_interval <= MaxPulseInterval)
                except:
                    logger.warning('ignoring invalid computation request')
                    client.send(None)
                    continue
                while 1:
                    computation._auth = Scheduler.auth_code()
                    if not os.path.exists(os.path.join(self.__dest_path, computation._auth)):
                        break
                try:
                    os.mkdir(os.path.join(self.__dest_path, computation._auth))
                except:
                    logger.debug('Could not create "%s"',
                                 os.path.join(self.__dest_path, computation._auth))
                    client.send(None)
                    continue
                # TODO: save it on disk instead
                computations[computation._auth] = computation
                client.send(computation._auth)

            elif req == 'await':
                if not isinstance(client, Coro):
                    logger.warning('Ignoring invalid client request "%s"', req)
                    continue
                computation = computations.pop(auth, None)
                if not computation:
                    client.send(None)
                    continue
                if computation._pulse_coro.location.addr != self.asyncoro.location.addr:
                    computation._xfer_files = [(os.path.join(self.__dest_path, computation._auth,
                                                             os.path.join(*(dst.split(sep))),
                                                             xf.split(sep)[-1]),
                                                os.path.join(*(dst.split(sep))), os.sep)
                                               for xf, dst, sep in computation._xfer_files]
                for xf, dst, sep in computation._xfer_files:
                    if not os.path.isfile(xf):
                        logger.warning('File "%s" for computation %s is not valid',
                                       xf, computation._auth)
                        computation = None
                        break
                if computation is None:
                    client.send(None)
                else:
                    self.__computation_scheduler_coro.send((computation, client))
                    self.__computation_sched_event.set()

            elif req == 'close_computation':
                if not isinstance(client, Coro):
                    logger.warning('Ignoring invalid client request "%s"', req)
                    continue
                SysCoro(self.__close_computation, client=client,
                        await_async=msg.get('await_async', False))

            else:
                logger.warning('Ignoring invalid client request "%s"', req)

    def __close_node(self, node, computation, await_async=False, coro=None):
        if (not computation) or (node.status not in (Scheduler.NodeInitialized,
                                                     Scheduler.NodeDiscovered)):
            logger.warning('Closing node %s ignored: %s', node.addr, node.status)
            raise StopIteration(-1)
        while node.cpus_used:
            logger.info('Waiting for %s remote coroutines at %s to finish',
                        node.cpus_used, node.addr)
            node.avail.clear()
            yield node.avail.wait()
        close_coros = [SysCoro(self.__close_server, server, computation, await_async=await_async)
                       for server in node.servers.values()]
        close_coros.extend([SysCoro(self.__close_server, server, computation)
                            for server in node.disabled_servers.values() if server.coro])
        for close_coro in close_coros:
            yield close_coro.finish()
        if node.coro:
            node.coro.send({'req': 'release', 'auth': computation._auth, 'client': coro})
            yield coro.receive(timeout=MsgTimeout)
        node.status = Scheduler.NodeClosed

    def __close_server(self, server, computation, await_async=False, coro=None):
        if ((not computation or not server.coro) or
            (server.status not in (Scheduler.ServerInitialized, Scheduler.ServerDiscovered))):
            logger.debug('Closing server %s ignored: %s', server.coro.location if server.coro else '',
                         server.status)
            raise StopIteration(-1)
        node = self._nodes.get(server.coro.location.addr, None)
        if not node:
            raise StopIteration(-1)
        server_coro, server.coro = server.coro, None
        if self.__server_locations:
            self.__server_locations.discard(server_coro.location)
            # TODO: inform other servers
        disconnected = node.servers.pop(server_coro.location, None) is None
        if disconnected:
            if computation and computation.status_coro:
                computation.status_coro.send(DiscoroStatus(Scheduler.ServerDisconnected,
                                                           server_coro.location))
        else:
            if not server.avail.is_set():
                logger.debug('Waiting for remote coroutines at %s to finish', server_coro.location)
                yield server.avail.wait()
            if await_async:
                while server.rcoros:
                    rcoro, job = server.rcoros[next(iter(server.rcoros))]
                    logger.debug('Waiting for %s to finish', rcoro)
                    yield job.done.wait()
            server_coro.send({'req': 'close', 'auth': computation._auth, 'client': coro})
            yield coro.receive(timeout=MsgTimeout)
        if server.rcoros:  # wait a bit for server to terminate coros
            for _ in range(5):
                yield coro.sleep(0.1)
                if not server.rcoros:
                    break
        if server.rcoros:
            logger.warning('%s coros running at %s', len(server.rcoros), server_coro.location)
            if computation and computation.status_coro:
                for (rcoro, cpu) in server.rcoros.values():
                    status = asyncoro.MonitorException(rcoro, (Scheduler.ServerClosed, None))
                    computation.status_coro.send(status)
                node.cpus_used -= len(server.rcoros)
                server.rcoros.clear()

        server.status = Scheduler.ServerClosed
        server.xfer_files = []
        server.askew_results.clear()
        if computation and computation.status_coro:
            computation.status_coro.send(DiscoroStatus(server.status, server_coro.location))

        if node.cpus_used == len(node.servers):
            self._avail_nodes.discard(node)
            if not self._avail_nodes:
                self._nodes_avail.clear()
            node.avail.clear()

        if disconnected and not node.servers:
            node.cpus_used = 0
            node.load = 0.0
            node.status = Scheduler.NodeClosed
            if computation and computation.status_coro:
                info = DiscoroNodeInfo(node.name, node.addr, node.cpus,
                                       node.platform, node.avail_info)
                computation.status_coro.send(DiscoroStatus(node.status, info))

        raise StopIteration(0)

    def __close_computation(self, client=None, await_async=False, coro=None):
        self.__server_locations.clear()
        if self._cur_computation:
            close_coros = [SysCoro(self.__close_node, node, self._cur_computation,
                                   await_async=await_async) for node in self._nodes.values()]
            close_coros.extend([SysCoro(self.__close_node, node, self._cur_computation)
                                for node in self._disabled_nodes.values()
                                if node.status == Scheduler.NodeDiscovered])
            for close_coro in close_coros:
                yield close_coro.finish()
        if self.__cur_client_auth:
            computation_path = os.path.join(self.__dest_path, self.__cur_client_auth)
            if os.path.isdir(computation_path):
                shutil.rmtree(computation_path, ignore_errors=True)
        if self._cur_computation and self._cur_computation.status_coro:
            self._cur_computation.status_coro.send(DiscoroStatus(Scheduler.ComputationClosed,
                                                                 id(self._cur_computation)))
        self.__cur_client_auth = self._cur_computation = None
        self.__computation_sched_event.set()
        if client:
            client.send('closed')
        raise StopIteration(0)

    def close(self, coro=None):
        """Close current computation and quit scheduler.

        Must be called with 'yield' as 'yield scheduler.close()' or as
        coroutine.
        """
        yield self.__close_computation(coro=coro)
        raise StopIteration(0)


if __name__ == '__main__':
    """The scheduler can be started either within a client program (if no other
    client programs use the nodes simultaneously), or can be run on a node with
    the options described below (usually no options are necessary, so the
    scheduler can be strated with just 'discoro.py')
    """

    import logging
    import argparse
    import signal
    try:
        import readline
    except ImportError:
        pass

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--ip_addr', dest='node', default=None,
                        help='IP address or host name of this node')
    parser.add_argument('--ext_ip_addr', dest='ext_ip_addr', default=None,
                        help='External IP address to use (needed in case of NAT firewall/gateway)')
    parser.add_argument('-u', '--udp_port', dest='udp_port', type=int, default=51350,
                        help='UDP port number to use')
    parser.add_argument('-t', '--tcp_port', dest='tcp_port', type=int, default=0,
                        help='TCP port number to use')
    parser.add_argument('-n', '--name', dest='name', default=None,
                        help='(symbolic) name given to schduler')
    parser.add_argument('--dest_path', dest='dest_path', default=None,
                        help='path prefix to where files sent by peers are stored')
    parser.add_argument('--max_file_size', dest='max_file_size', default=None, type=int,
                        help='maximum file size of any file transferred')
    parser.add_argument('-s', '--secret', dest='secret', default='',
                        help='authentication secret for handshake with peers')
    parser.add_argument('--certfile', dest='certfile', default='',
                        help='file containing SSL certificate')
    parser.add_argument('--keyfile', dest='keyfile', default='',
                        help='file containing SSL key')
    parser.add_argument('--node', action='append', dest='nodes', default=[],
                        help='additional remote nodes (names or IP address) to use')
    parser.add_argument('--pulse_interval', dest='pulse_interval', type=float,
                        default=MaxPulseInterval,
                        help='interval in seconds to send "pulse" messages to check nodes '
                        'and client are connected')
    parser.add_argument('--ping_interval', dest='ping_interval', type=float, default=0,
                        help='interval in seconds to broadcast "ping" message to discover nodes')
    parser.add_argument('--zombie_period', dest='zombie_period', type=int,
                        default=(100 * MaxPulseInterval),
                        help='maximum time in seconds computation is idle')
    parser.add_argument('-d', '--debug', action='store_true', dest='loglevel', default=False,
                        help='if given, debug messages are printed')
    parser.add_argument('--clean', action='store_true', dest='clean', default=False,
                        help='if given, files copied from or generated by clients will be removed')
    parser.add_argument('--discoronode_port', dest='discoronode_port', type=int, default=51351,
                        help='UDP port number used by discoronode')
    parser.add_argument('--daemon', action='store_true', dest='daemon', default=False,
                        help='if given, input is not read from terminal')
    config = vars(parser.parse_args(sys.argv[1:]))
    del parser

    if config['zombie_period'] and config['zombie_period'] < MaxPulseInterval:
        raise Exception('zombie_period must be >= %s' % MaxPulseInterval)

    if not config['name']:
        config['name'] = 'discoro_scheduler'

    if config['loglevel']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    del config['loglevel']

    if config['certfile']:
        config['certfile'] = os.path.abspath(config['certfile'])
    else:
        config['certfile'] = None
    if config['keyfile']:
        config['keyfile'] = os.path.abspath(config['keyfile'])
    else:
        config['keyfile'] = None

    daemon = config.pop('daemon', False)
    _discoro_scheduler = Scheduler(**config)
    _discoro_scheduler._shared = True
    del config

    def sighandler(signum, frame):
        # Coro(_discoro_scheduler.close).value()
        raise KeyboardInterrupt

    try:
        signal.signal(signal.SIGHUP, sighandler)
        signal.signal(signal.SIGQUIT, sighandler)
    except:
        pass
    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGABRT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)
    del sighandler

    if not daemon:
        try:
            if os.getpgrp() != os.tcgetpgrp(sys.stdin.fileno()):
                daemon = True
        except:
            pass

    if daemon:
        del daemon
        while 1:
            try:
                time.sleep(3600)
            except:
                break
    else:
        del daemon
        while 1:
            try:
                _discoro_cmd = input(
                    '\n\nEnter "quit" or "exit" to terminate discoro scheduler\n'
                    '      "status" to show status of scheduler: '
                    )
            except:
                break
            _discoro_cmd = _discoro_cmd.strip().lower()
            if _discoro_cmd in ('quit', 'exit'):
                break
            if _discoro_cmd == 'status':
                _discoro_scheduler.print_status()

    logger.info('terminating discoro scheduler')
    Coro(_discoro_scheduler.close).value()



# Run 'discoronode.py' program to start processes to execute computations sent
# by this client, along with this program.

# This example uses status messages and message passing to run two remote
# coroutines at two different servers to process streaming data for
# live/real-time analysis. This example requires numpy module available on
# servers. See discoro_client6_channel.py that uses channels for communication
# and 'deque' for circular buffers (instead of numpy).

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *


# This generator function is sent to remote discoro to analyze data
# and generate apprporiate signals that are sent to a coroutine
# running on client. The signal in this simple case is average of
# moving window of given size is below or above a threshold.
def rcoro_avg_proc(threshold, trend_coro, window_size, coro=None):
    import numpy as np
    data = np.empty(window_size, dtype=float)
    data.fill(0.0)
    cumsum = 0.0
    while True:
        i, n = yield coro.receive()
        if n is None:
            break
        cumsum += (n - data[0])
        avg = cumsum / window_size
        if avg > threshold:
            trend_coro.send((i, 'high', float(avg)))
        elif avg < -threshold:
            trend_coro.send((i, 'low', float(avg)))
        data = np.roll(data, -1)
        data[-1] = n
    raise StopIteration(0)


# This generator function is sent to remote discoro process to save
# the received data in a file (on the remote peer).
def rcoro_save_proc(coro=None):
    import os
    import tempfile
    # save data in /tmp/tickdata
    with open(os.path.join(os.sep, tempfile.gettempdir(), 'tickdata'), 'w') as fd:
        while True:
            i, n = yield coro.receive()
            if n is None:
                break
            fd.write('%s: %s\n' % (i, n))
    raise StopIteration(0)


# This coroutine runs on client. It gets trend messages from remote
# coroutine that computes moving window average.
def trend_proc(coro=None):
    coro.set_daemon()
    while True:
        trend = yield coro.receive()
        print('trend signal at % 4d: %s / %.2f' % (trend[0], trend[1], trend[2]))


# This process runs locally. It creates two remote coroutines at two discoronode
# server processes, two local coroutines, one to receive trend signal from one
# of the remote coroutines, and another to send data to two remote coroutines
def client_proc(computation, coro=None):
    # schedule computation with the scheduler; scheduler accepts one computation
    # at a time, so if scheduler is shared, the computation is queued until it
    # is done with already scheduled computations
    if (yield computation.schedule()):
        raise Exception('Could not schedule computation')

    trend_coro = asyncoro.Coro(trend_proc)

    # run average and save coroutines at two different servers
    rcoro_avg = yield computation.run(rcoro_avg_proc, 0.4, trend_coro, 10)
    assert isinstance(rcoro_avg, asyncoro.Coro)
    rcoro_save = yield computation.run(rcoro_save_proc)
    assert isinstance(rcoro_save, asyncoro.Coro)

    # if data is sent frequently (say, many times a second), enable
    # streaming data to remote peer; this is more efficient as
    # connections are kept open (so the cost of opening and closing
    # connections is avoided), but keeping too many connections open
    # consumes system resources
    yield asyncoro.AsynCoro.instance().peer(rcoro_avg.location, stream_send=True)
    yield asyncoro.AsynCoro.instance().peer(rcoro_save.location, stream_send=True)

    # send 1000 items of random data to remote coroutines
    for i in range(1000):
        n = random.uniform(-1, 1)
        item = (i, n)
        # data can be sent to remote coroutines either with 'send' or
        # 'deliver'; 'send' is more efficient but no guarantee data
        # has been sent successfully whereas 'deliver' indicates
        # errors right away; alternately, messages can be sent with a
        # channel, which is more convenient if there are multiple
        # (unknown) recipients
        rcoro_avg.send(item)
        rcoro_save.send(item)
        yield coro.sleep(0.01)
    item = (i, None)
    rcoro_avg.send(item)
    rcoro_save.send(item)

    yield computation.close()


if __name__ == '__main__':
    import random
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is shared (i.e., running as program), nothing needs
    # to be done (its location can optionally be given to 'schedule');
    # othrwise, start private scheduler:
    Scheduler()
    computation = Computation([])
    asyncoro.Coro(client_proc, computation)
